import numpy as np
import scipy.stats as stats
import scipy.interpolate as interp
from copy import deepcopy
from typing import List, Dict, Union

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.spectrumData import SpectrumData
from resistics.utilities.utilsPrint import listToString, arrayToString
from resistics.utilities.utilsRobust import olsModel
from resistics.utilities.utilsSmooth import smooth1d


class StatisticCalculator(Calculator):
    """Calculate statistics for data restriction

    Statistics are calculated out for each evaluation frequency in each window. Therefore, there are nwindow*nfreq statistics in total.

    This class was written to speed up statistic calculations. Many statistics need the same data, for example power spectra. This class calculates and reuses some common values amongst the various statistics to improve calculation speed.

    Attributes
    ----------
    evalFreq : List
        List of evaluation frequencies
    winLen : int
        Window length for spectra calculations
    winType : str
        Window function to apply to time data before fourier transform
    inChans : List[str]
        Input channels
    inSize : int
        Number of input channels
    outChans : List[str]
        Output channels 
    outSize : int
        Number of output channels
    specChans : List[str] 
        The channels for which to calculate auto and cross power spectra
    remoteChans : List[str]
        Remote reference channels
    psdChans : List[str] 
        Power spectral density channels
    cohPairs : List[List[str]]
        Pairs of channels for coherence calculations
    polDirs : List[List[str]]
        Pairs of channels of polarisation direction calculation
    spec: Dict = {}

    tfCalculated : bool 
        Boolean flag to show that the transfer function has been calculated for a window
    remoteCalculated : bool 
        Boolean flag to show...
    intercept : bool   
        Boolean flag to include an intercept into the 
    outData: Dict = {}

    Methods
    -------
    __init__()
    getEvalFreq()
    getInChans()
    getOutChans()
    getSpecChans()
    getRemoteChans()
    getPSDChans()
    getCohPairs()
    getPolDirs()
    getAutoPower(chan)
    getAutoPowerEval(chan, eIdx)
    getCrossPower(chan1, chan2)
    getCrossPowerEval(chan1, chan2, eIdx)
    getOutData()
    setInChans(inChans)
    setOutChans(outChans)
    setSpectra(freq, winData, evalFreq)
    setIntercept(intercept)
    calculateSpectralMatrix()
    calculateEvalMatrix()
    addRemoteSpec(remoteData, **kwargs)
    calculateRemoteSpectralMatrix()
    calculateRemoteEvalMatrix()
    calculateReferenceSpectralMatrix()
    calculateReferenceEvalMatrix()
    getRemoteAutoPower(chan)
    getRemoteAutoPowerEval(chan, eIdx)
    getRemoteCrossPower(chan1, chan2)
    getRemoteCrossPowerEval(chan1, chan2, eIdx)
    getReferenceCrossPower(dataChan, remoteChan)
    getReferenceCrossPowerEval(dataChan, remoteChan, eIdx)
    interpolateToEvalFreq(data)
    prepareOutDict()
    getDataForStatName(statName)
    winPSD()
    winCoherence()
    winPolarisations()
    winPartials()
    winTransferFunction()
    winRemoteCoherence()
    winRemoteEqnCoherence()
    winRemoteAbsVal()
    winRemoteTransferFunction()
    printList()
        Class information returned as list of strings  
    """

    def __init__(self) -> None:
        """Initialise the statistic calculator"""

        # default evaluation frequencies
        self.evalFreq: List = []
        # power smoothing vals
        self.winLen: int = 13
        self.winType: str = "hanning"
        # set some defaults
        self.inChans: List[str] = ["Hx", "Hy"]
        self.inSize: int = len(self.inChans)
        self.outChans: List[str] = ["Ex", "Ey"]
        self.outSize: int = len(self.outChans)
        self.specChans: List[str] = self.inChans + self.outChans
        self.remoteChans: List[str] = self.inChans
        self.psdChans: List[str] = ["Ex", "Ey", "Hx", "Hy"]
        self.cohPairs: List[List[str]] = [
            ["Ex", "Hx"],
            ["Ex", "Hy"],
            ["Ey", "Hx"],
            ["Ey", "Hy"],
        ]
        self.polDirs: List[List[str]] = [["Ex", "Ey"], ["Hx", "Hy"]]
        # set data presets
        self.freq: Union[np.ndarray, None] = None
        self.spec: Dict[np.ndarray] = {}
        # output data and marker for transfer function calculated
        self.tfCalculated: bool = False
        self.remoteCalculated: bool = False
        self.intercept: bool = False
        self.outData: Dict = {}

    def getEvalFreq(self):
        """Get a copy of the evaluation frequency
        
        Returns
        -------
        List[float]
            List of evaluation frequencies
        """

        return deepcopy(self.evalFreq)

    def getInChans(self) -> List[str]:
        """Get a copy of the input channels
        
        Returns
        -------
        List[str]
            List of input channels
        """

        return deepcopy(self.inChans)

    def getOutChans(self) -> List[str]:
        """Get a copy of the output channels
        
        Returns
        -------
        List[str]
            List of output channels
        """

        return deepcopy(self.outChans)

    def getSpecChans(self):
        """Get a copy of the channels for which to calculate spectra
        
        Returns
        -------
        List[str]
            List of spectra channels
        """

        return deepcopy(self.specChans)

    def getRemoteChans(self) -> List[str]:
        """Get a copy of the remote reference channels
        
        Returns
        -------
        List[str]
            List of remote reference channels
        """

        return deepcopy(self.remoteChans)

    def getPSDChans(self) -> List[str]:
        """Get a copy of the channels to include power spectral density
        
        Returns
        -------
        List[str]
            List of power spectral density channels
        """

        return deepcopy(self.psdChans)

    def getCohPairs(self) -> List[List[str]]:
        """Get a copy of coherence channel pairs to calculate out
        
        Returns
        -------
        List[List[str]]
            List of coherence pairs
        """

        return deepcopy(self.cohPairs)

    def getPolDirs(self) -> List[List[str]]:
        """Get a list of polarisation direction pairs
        
        Returns
        -------
        List[List[str]]
            List of polarisation direction pairs
        """

        return deepcopy(self.polDirs)

    def getAutoPower(self, chan: str) -> np.ndarray:
        """Get the auto power for a channel

        Parameters
        ----------
        chan : str
            The channel for which to get the autopower
        
        Returns
        -------
        np.ndarray
            The auto power for the channel
        """

        idx = self.specChans.index(chan)
        # then return the autopower
        return self.spectralMatrix[idx, idx].real

    def getAutoPowerEval(self, chan: str, eIdx: int) -> float:
        """Get the auto power value for an particular evaluation frequency

        Parameters
        ----------
        chan : str
            The channel for which to get the autopower
        eIdx : int
            The index for the evaluation frequency
        
        Returns
        -------
        np.ndarray
            The auto power for the channel
        """

        idx = self.specChans.index(chan)
        # then return the autopower
        return self.evalMatrix[idx, idx, eIdx].real

    def getCrossPower(self, chan1: str, chan2: str) -> np.ndarray:
        """Get the cross power between two channels

        Parameters
        ----------
        chan1 : str
            The first channel for the cross channels
        chan2 : str
            The second channl for the cross channels
        
        Returns
        -------
        np.ndarray
            The cross power spectral density
        """

        idx1 = self.specChans.index(chan1)
        idx2 = self.specChans.index(chan2)
        # then return the autopower
        return self.spectralMatrix[idx1, idx2]

    def getCrossPowerEval(self, chan1: str, chan2: str, eIdx: int) -> float:
        """Get the cross power between two channels

        Parameters
        ----------
        chan1 : str
            The first channel for the cross channels
        chan2 : str
            The second channl for the cross channels
        eIdex : int
            The index of the evaluation frequency
        
        Returns
        -------
        np.ndarray
            The cross power spectral density
        """

        idx1 = self.specChans.index(chan1)
        idx2 = self.specChans.index(chan2)
        # then return the autopower
        return self.evalMatrix[idx1, idx2, eIdx]

    def getOutData(self) -> Dict:
        """Get the output data

        Returns
        -------
        Dict
            The statistic output data
        """

        return deepcopy(self.outData)

    def setInChans(self, inChans: List[str]) -> None:
        """Set the input channels

        Parameters
        ----------
        inChans : List[str]
            Input channels for the magnetotelluric linear system
        """

        self.inChans = inChans
        self.inSize = len(self.inChans)

    def setOutChans(self, outChans: List[str]):
        """Set the output channels

        Parameters
        ----------
        inChans : List[str]
            Output channels for the magnetotelluric linear system
        """

        self.outChans = outChans
        self.outSize = len(self.outChans)

    def setRemoteChans(self, remoteChans):
        """Set the input channels

        Parameters
        ----------
        inChans : List[str]
            Input channels for the magnetotelluric linear system
        """

        self.remoteChans = remoteChans

    def setPSDChans(self, psdChans: List[str]) -> None:
        """Set the power spectral density channels

        Parameters
        ----------
        psdChans : List[str]
            Power spectral density channels. An example input would be: ["Ex", "Ey", "Hx", "Hy"]
        """

        self.psdChans = psdChans

    def setCohPairs(self, cohPairs: List[List[str]]) -> None:
        """Set the power spectral density channels

        If cohPairs of [["Ex", "Hx"], ["Ex", "Hy"], ["Ey", "Hx"], ["Ey", "Hy"]] are set, the following coherences will be calculated:
        ExHx
        ExHy
        EyHx
        EyHy

        Parameters
        ----------
        cohPairs : List[List[str]]
            Set the coherence pairs using a list of channel pairs, for example: [["Ex", "Hx"], ["Ex", "Hy"], ["Ey", "Hx"], ["Ey", "Hy"]]
        """

        self.cohPairs = cohPairs

    def setPolDirs(self, polDirs: List[List[str]]) -> None:
        """Set the polarisation direction pairs to calculate

        If polDirs of [["Ex", "Ey"], ["Hx", "Hy"]] are set, the following polarisation directions will be calculated:
        Ex Ey
        Hx Hy

        Parameters
        ----------
        polDirs : List[List[str]]
            Set polarisation direction channel pairs, for example: [["Ex", "Ey"], ["Hx", "Hy"]]
        """

        self.polDirs = polDirs

    def setSpectra(
        self, freq: np.ndarray, specData: SpectrumData, evalFreq: np.ndarray
    ) -> None:
        """Set the spectra data

        Parameters
        ----------
        freq : np.ndarray
            The frequency points in the spectra data
        specData : SpectrumData
            Spectrum data (i.e. spectrum data for a window)
        evalFreq : np.ndarray
            Evaluation frequency array
        """

        self.freq = freq
        self.spec: Dict[np.ndarray] = specData.data
        self.evalFreq = evalFreq
        # self.specChans = sorted(self.spec.keys())
        self.numChans = len(self.specChans)
        self.dataSize = specData.dataSize
        # calculate the power matrix
        self.calculateSpectralMatrix()
        self.calculateEvalMatrix()
        # clear the out dictionary and set that transfer function not calculated
        self.prepareOutDict()

    def setIntercept(self, intercept: bool):
        """Set the intercept boolean

        Parameters
        ----------
        intercept : bool
            Boolean flag for having an intercept in the transfer function calculation
        """

        self.intercept = intercept

    def calculateSpectralMatrix(self) -> None:
        """Calculate out the cross power spectral matrix

        The method calculates out the cross powers which will then be used in the other statistic calculations.

        The cross powers spectral matrix is a 3-D matrix of size:
        numChans * numChans * numFrequencies
        The elements of this are calculated by multiplying the spectra of one channel by the complex conjugate of the spectra of another channel.
        """

        # create the 3d array
        self.spectralMatrix = np.empty(
            shape=(self.numChans, self.numChans, self.dataSize), dtype="complex"
        )
        # now need to go through the chans
        for ii in range(0, self.numChans):
            for jj in range(ii, self.numChans):
                chan1 = self.specChans[ii]
                chan2 = self.specChans[jj]
                self.spectralMatrix[ii, jj] = smooth1d(
                    self.spec[chan1] * np.conjugate(self.spec[chan2]),
                    self.winLen,
                    self.winType,
                )
                if ii == jj:
                    # conjugate symmtry
                    self.spectralMatrix[jj, ii] = np.conjugate(
                        self.spectralMatrix[ii, jj]
                    )

    def calculateEvalMatrix(self):
        """Calculate out the cross power spectral matrix at the evaluation frequencies

        The method calculates out the cross powers which will then be used in the other statistic calculations at the evaluation frequencies

        The cross powers spectral matrix for evaluation frequencies is a 3-D matrix of size:
        numChans * numChans * numEvaluationFrequencies
        The elements of this are calculated by taking the cross powers spectral matrix and using the result there to interpolate the values at the evaluation frequencies.
        """

        # create the array
        self.evalMatrix = np.empty(
            shape=(self.numChans, self.numChans, len(self.evalFreq)), dtype="complex"
        )
        for ii in range(0, self.numChans):
            for jj in range(ii, self.numChans):
                self.evalMatrix[ii, jj] = self.interpolateToEvalFreq(
                    self.spectralMatrix[ii, jj]
                )
                if ii != jj:
                    # conjugate symmtry
                    self.evalMatrix[jj, ii] = np.conjugate(self.evalMatrix[ii, jj])

    def addRemoteSpec(self, remoteData: SpectrumData, remoteChans: List[str] = []) -> None:
        """Add coincident remote reference spectrum data

        Parameters
        ----------
        remoteData : SpectrumData
            Spectrum data (i.e. spectrum data for a window)
        remoteChans : List[str]
            The channels to use from remote reference data
        """

        self.remoteSpec = remoteData.data
        if len(remoteChans) > 0:
            self.remoteChans = remoteChans
        # now calculate some remote reference related values
        self.calculateRemoteSpectralMatrix()
        self.calculateRemoteEvalMatrix()
        self.calculateReferenceSpectralMatrix()
        self.calculateReferenceEvalMatrix()

    def calculateRemoteSpectralMatrix(self):
        """Calculate out the cross power spectral matrix for the remote reference data

        The method calculates out the cross powers for the remote reference channels which will then be used in the other statistic calculations.

        The remote reference cross powers spectral matrix is a 3-D matrix of size:
        numRemoteChans * numRemoteChans * numFrequencies
        The elements of this are calculated by multiplying the spectra of one channel by the complex conjugate of the spectra of another channel.
        """

        # create the 3d array
        numRemoteChans = len(self.remoteChans)
        self.remoteSpectralMatrix = np.empty(
            shape=(numRemoteChans, numRemoteChans, self.dataSize), dtype="complex"
        )
        # now need to go through the chans
        for ii in range(0, numRemoteChans):
            for jj in range(ii, numRemoteChans):
                chan1 = self.remoteChans[ii]
                chan2 = self.remoteChans[jj]
                self.remoteSpectralMatrix[ii, jj] = smooth1d(
                    self.remoteSpec[chan1] * np.conjugate(self.remoteSpec[chan2]),
                    self.winLen,
                    self.winType,
                )
                if ii == jj:
                    # conjugate symmtry
                    self.remoteSpectralMatrix[jj, ii] = np.conjugate(
                        self.remoteSpectralMatrix[ii, jj]
                    )

    def calculateRemoteEvalMatrix(self):
        """Calculate out the cross power spectral matrix for the remote reference data at the evaluation frequencies

        Takes the cross power spectral data calculate for the remote reference channels and interpoaltes it to the evaluation frequencies
        """

        # create the array
        numRemoteChans = len(self.remoteChans)
        self.remoteEvalMatrix = np.empty(
            shape=(numRemoteChans, numRemoteChans, len(self.evalFreq)), dtype="complex"
        )
        for ii in range(0, numRemoteChans):
            for jj in range(ii, numRemoteChans):
                self.remoteEvalMatrix[ii, jj] = self.interpolateToEvalFreq(
                    self.remoteSpectralMatrix[ii, jj]
                )
                if ii != jj:
                    # conjugate symmtry
                    self.remoteEvalMatrix[jj, ii] = np.conjugate(
                        self.remoteEvalMatrix[ii, jj]
                    )

    def calculateReferenceSpectralMatrix(self):
        """Calculate out the cross power spectral matrix between the site spectral data and the remote reference spectral data

        The reference cross powers spectral matrix is a 3-D matrix of size:
        numChans * numRemoteChans * numFrequencies
        The elements of this are calculated by multiplying a channel of the site spectral data by the complex conjugate of a channel from the remote reference. 
        """

        # cannot use conjugate symmetry in this case
        self.referenceSpectralMatrix = np.empty(
            shape=(self.numChans, len(self.remoteChans), self.dataSize), dtype="complex"
        )
        for ii, chan1 in enumerate(self.specChans):
            for jj, chan2 in enumerate(self.remoteChans):
                self.referenceSpectralMatrix[ii, jj] = smooth1d(
                    self.spec[chan1] * np.conjugate(self.remoteSpec[chan2]),
                    self.winLen,
                    self.winType,
                )

    def calculateReferenceEvalMatrix(self):
        """Interpolate the remote and site cross powers spectral matrix to the evaluation frequencies.
        """

        self.referenceEvalMatrix = np.empty(
            shape=(self.numChans, len(self.remoteChans), len(self.evalFreq)),
            dtype="complex",
        )
        for ii, chan1 in enumerate(self.specChans):
            for jj, chan2 in enumerate(self.remoteChans):
                self.referenceEvalMatrix[ii, jj] = self.interpolateToEvalFreq(
                    self.referenceSpectralMatrix[ii, jj]
                )

    def getRemoteAutoPower(self, chan: str) -> np.ndarray:
        """Get the auto power of a remote reference channel

        Parameters
        ----------
        chan : str
            The channel for which to get the autopower
        
        Returns
        -------
        np.ndarray
            The autopower array (real for autopowers)
        """

        idx = self.remoteChans.index(chan)
        return self.remoteSpectralMatrix[idx, idx].real

    def getRemoteAutoPowerEval(self, chan: str, eIdx: int) -> float:
        """Get the auto power of a remote reference channel at an evaluation frequency

        Parameters
        ----------
        chan : str
            The channel for which to get the autopower
        eIdx : int
            The evaluation frequency index
        
        Returns
        -------
        float
            The autopower of the channel at the evaluation frequency
        """

        idx = self.remoteChans.index(chan)
        return self.remoteEvalMatrix[idx, idx, eIdx].real

    def getRemoteCrossPower(self, chan1: str, chan2: str) -> np.ndarray:
        """Get the cross power of two remote reference channels

        Parameters
        ----------
        chan1 : str
            The first channel for the cross power
        chan2 : str
            The second channel for the cross power
        
        Returns
        -------
        np.ndarray
            The cross power array 
        """

        idx1 = self.remoteChans.index(chan1)
        idx2 = self.remoteChans.index(chan2)
        return self.remoteSpectralMatrix[idx1, idx2]

    def getRemoteCrossPowerEval(self, chan1: str, chan2: str, eIdx: int) -> float:
        """Get the cross power of two remote reference channels at a single evaluation frequency

        Parameters
        ----------
        chan1 : str
            The first channel for the cross power
        chan2 : str
            The second channel for the cross power
        eIdx : int
            The evaluation frequency index            
        
        Returns
        -------
        float
            The value of the cross power at the evaluation frequency 
        """

        idx1 = self.remoteChans.index(chan1)
        idx2 = self.remoteChans.index(chan2)
        return self.remoteSpectralMatrix[idx1, idx2, eIdx]

    def getReferenceCrossPower(self, dataChan: str, remoteChan: str) -> np.ndarray:
        """Get the cross power of a data channel and a remote reference channel

        Parameters
        ----------
        dataChan : str
            The data channel
        remoteChan : str
            The remote reference channel
        
        Returns
        -------
        np.ndarray
            The cross power array 
        """

        idx1 = self.specChans.index(dataChan)
        idx2 = self.remoteChans.index(remoteChan)
        return self.referenceSpectralMatrix[idx1, idx2]

    def getReferenceCrossPowerEval(
        self, dataChan: str, remoteChan: str, eIdx: int
    ) -> float:
        """Get the cross power of a data channel and a remote reference channel at a single evaluation frequency

        Parameters
        ----------
        dataChan : str
            The data channel
        remoteChan : str
            The remote reference channel
        eIdx : int
            The evaluation frequency index              
        
        Returns
        -------
        float
            The value of the cross power at the evaluation frequency 
        """

        idx1 = self.specChans.index(dataChan)
        idx2 = self.remoteChans.index(remoteChan)
        return self.referenceEvalMatrix[idx1, idx2, eIdx]

    def interpolateToEvalFreq(self, data: np.ndarray) -> np.ndarray:
        """Interpolate data on to the evaluation frequency

        Parameters
        ----------
        data : np.ndarray
            Power spectral data defined at frequency points given in the freqs array

        Returns
        -------
        np.ndarray
            Data interpolated to evaluation frequencies
        """

        interpFunc = interp.interp1d(self.freq, data)
        interpData = interpFunc(self.evalFreq)
        return interpData

    def prepareOutDict(self) -> None:
        """Prepare output statistic output dictionary
        
        The outData dictionary is indexed in the following way:
        outData[evaluation frequency][statistic component] = value
        """

        self.outData = {}
        for e in self.evalFreq:
            self.outData[e] = {}
        # set various calculated flags to false
        self.tfCalculated = False
        self.remoteCalculated = False

    def getDataForStatName(self, statName: str):
        """Return the data for a statistic

        Given a statitic name, this method returns data from the correct internal method.

        Parameters
        ----------
        statName : str
            The name of the statistic to calculate out
        
        Returns
        -------
        Dict
            The output dictionary
        """

        if statName == "absvalEqn":
            return self.winAbsVal()
        elif statName == "coherence":
            return self.winCoherence()
        elif statName == "powerSpectralDensity":
            return self.winPSD()
        elif statName == "polarisationDirection":
            return self.winPolarisations()
        elif statName == "partialCoherence":
            return self.winPartials()
        elif statName == "transferFunction" or statName == "resPhase":
            if self.tfCalculated:
                return self.getOutData()
            return self.winTransferFunction()
        elif statName == "RR_coherence":
            return self.winRemoteCoherence()
        elif statName == "RR_coherenceEqn":
            return self.winRemoteEqnCoherence()
        elif statName == "RR_absvalEqn":
            return self.winRemoteAbsVal()
        elif statName == "RR_transferFunction" or statName == "RR_resPhase":
            if self.remoteCalculated:
                return self.getOutData()
            return self.winRemoteTransferFunction()
        else:
            self.printError(
                "Statistic in getDataForStatName not recognised", quitRun=True
            )
            return self.winCoherence()

    def winPSD(self):
        """Calculate power spectral densities

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        # need to divide by length of time too
        freqLen = self.freq.size
        timeLen = (freqLen - 1) * 2  # minus 1 because time sections are usually even
        fs = self.freq[-1] * 2  # sampling frequency
        # and then calculate amount of time
        duration = timeLen / fs
        # interpolate onto evaluation frequency and output to outData
        for eIdx, eF in enumerate(self.evalFreq):
            for chan in self.getPSDChans():
                key = "psd{}".format(chan)
                self.outData[eF][key] = self.getAutoPowerEval(chan, eIdx) / duration
        return self.getOutData()

    def winCoherence(self):
        """Calculate spectral coherence pairs

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        for idx, p in enumerate(self.getCohPairs()):
            c1 = p[0]  # chan1
            c2 = p[1]  # chan2
            for eIdx, eF in enumerate(self.evalFreq):
                # calculate the nominator and denominator
                cohNom = np.power(
                    np.absolute(self.getCrossPowerEval(c1, c2, eIdx)), 2
                ).real
                cohDenom = self.getAutoPowerEval(c1, eIdx) * self.getAutoPowerEval(
                    c2, eIdx
                )
                # save in outData
                key = "coh{}".format(c1 + c2)
                self.outData[eF][key] = cohNom / cohDenom
        return self.getOutData()

    def winPolarisations(self):
        """Calculate polarisation directions

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        for idx, p in enumerate(self.getPolDirs()):
            c1 = p[0]  # chan1
            c2 = p[1]  # chan2
            for eIdx, eF in enumerate(self.evalFreq):
                # now calculate the nominator and denominator
                cohNom = (
                    2 * self.getCrossPowerEval(c1, c2, eIdx).real
                )  # take the real part of this
                cohDenom = self.getAutoPowerEval(c1, eIdx) - self.getAutoPowerEval(
                    c2, eIdx
                )
                # save to out dictionary
                key = "pol{}".format(c1 + c2)
                self.outData[eF][key] = np.arctan(cohNom / cohDenom) * (180.0 / np.pi)
        return self.getOutData()

    def winPartials(self):
        """Calculate partial coherencies

        Based on paper Weckmann, Magunia Ritter 2005.
        e.g. coherence Ex, Hx w.r.t Hy
        This currently only works for impedance tensor calculations and higher power partial coherencies are not supported.

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]

        Notes
        -----
        Based on paper by Weckmann, Magunia Ritter 2005.
        """

        # get the coherences - these will be required later
        winCoherence = self.winCoherence()

        for i, outChan in enumerate(self.outChans):
            for eIdx, eFreq in enumerate(self.evalFreq):
                inChan1 = self.inChans[0]
                inChan2 = self.inChans[1]
                xOutIn1 = self.getCrossPowerEval(outChan, inChan1, eIdx)
                xOutIn2 = self.getCrossPowerEval(outChan, inChan2, eIdx)
                xIn1In2 = self.getCrossPowerEval(inChan1, inChan2, eIdx)
                xIn2In1 = self.getCrossPowerEval(inChan2, inChan1, eIdx)
                # calculate out transFunc components
                denom = (
                    self.getAutoPowerEval(inChan1, eIdx)
                    * self.getAutoPowerEval(inChan2, eIdx)
                    - xIn1In2 * xIn2In1
                )
                # Z1
                Z1nom = (
                    xOutIn1 * self.getAutoPowerEval(inChan2, eIdx) - xIn2In1 * xOutIn2
                )
                Z1 = Z1nom / denom
                # Z2
                Z2nom = (
                    self.getAutoPowerEval(inChan1, eIdx) * xOutIn2 - xIn1In2 * xOutIn1
                )
                Z2 = Z2nom / denom
                # calculate bivariate coherency
                rb = Z1 * self.getCrossPowerEval(
                    inChan1, outChan, eIdx
                ) + Z2 * self.getCrossPowerEval(inChan2, outChan, eIdx)
                rb = rb / self.getAutoPowerEval(outChan, eIdx)
                # now calculate out partials
                # calculate partial inChan, outChan1 with respect to outChan2
                cohkey = "coh{}".format(outChan + inChan2)
                rp1 = (rb - winCoherence[eFreq][cohkey]) / (
                    1.0 - winCoherence[eFreq][cohkey]
                )
                # calculate partial inChan, outChan2 with respect to outChan1
                cohkey = "coh{}".format(outChan + inChan1)
                rp2 = (rb - winCoherence[eFreq][cohkey]) / (
                    1.0 - winCoherence[eFreq][cohkey]
                )
                # now save in outDict
                self.outData[eFreq]["bivar{}".format(outChan)] = rb
                self.outData[eFreq]["par{}".format(outChan + inChan1)] = rp1
                self.outData[eFreq]["par{}".format(outChan + inChan2)] = rp2
        return self.getOutData()

    def winAbsVal(self):
        """Absolute values of the cross power spectral matrix

        This data is often useful for cross plotting

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        for eIdx, eFreq in enumerate(self.evalFreq):
            for iChan, chan in enumerate(self.specChans):
                # first do the outchans multiplied by every other channel
                for iOut, outChan in enumerate(self.outChans):
                    absval = np.absolute(self.getCrossPowerEval(outChan, chan, eIdx))
                    key = "abs{}{}".format(outChan, chan)
                    self.outData[eFreq][key] = absval

                # then  do the inchans multiplied by every other channel
                for iIn, inChan in enumerate(self.inChans):
                    absval = np.absolute(self.getCrossPowerEval(inChan, chan, eIdx))
                    key = "abs{}{}".format(inChan, chan)
                    self.outData[eFreq][key] = absval
        # return the dictionary
        return self.getOutData()

    def winTransferFunction(self):
        """Calculate transfer function for the spectral data

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        totalSize = self.inSize + self.outSize
        # now want to calculate the transfer function for each evaluation frequency
        output = np.empty(
            shape=(self.evalFreq.size, self.outSize, self.inSize), dtype="complex"
        )
        for eIdx, eFreq in enumerate(self.evalFreq):
            # solve transfer function
            obs = np.empty(shape=(self.outSize, totalSize), dtype="complex")
            reg = np.empty(
                shape=(self.outSize, totalSize, self.inSize), dtype="complex"
            )
            for i in range(0, self.outSize):
                for j in range(0, totalSize):
                    # this is the observation row where,i is the observed output
                    # idx in the evaluation frequency
                    obs[i, j] = self.getCrossPowerEval(
                        self.outChans[i], self.specChans[j], eIdx
                    )
                    for k in range(0, self.inSize):
                        reg[i, j, k] = self.getCrossPowerEval(
                            self.inChans[k], self.specChans[j], eIdx
                        )

            for i in range(0, self.outSize):
                observation = obs[i, :]
                predictors = reg[i, :, :]
                # now do the solution
                out, resids, squareResid, rank, s = olsModel(
                    predictors, observation, intercept=self.intercept
                )
                # out, resids, scale, weights	= mmestimateModel(predictors, observation, intercept=False)
                # not interested in the intercept (const) term
                if self.intercept:
                    output[eIdx, i] = out[1:]
                else:
                    output[eIdx, i] = out

            # calculate components of transfer function and res and phase
            for i in range(0, self.outSize):
                for j in range(0, self.inSize):
                    period = 1.0 / eFreq
                    res = 0.2 * period * np.power(np.absolute(output[eIdx, i, j]), 2)
                    phase = np.angle(output[eIdx, i, j], deg=True)
                    keyRes = self.outChans[i] + self.inChans[j] + "Res"
                    keyPhase = self.outChans[i] + self.inChans[j] + "Phase"
                    self.outData[eFreq][keyRes] = res
                    self.outData[eFreq][keyPhase] = phase
                    # add the components
                    keyReal = self.outChans[i] + self.inChans[j] + "Real"
                    keyImag = self.outChans[i] + self.inChans[j] + "Imag"
                    self.outData[eFreq][keyReal] = output[eIdx, i, j].real
                    self.outData[eFreq][keyImag] = output[eIdx, i, j].imag
        # set transfer function calculated as true
        # saves having to do it again
        self.tfCalculated = True
        return self.getOutData()

    def winRemoteCoherence(self):
        """Calulate coherence between data channels and remote channels

        For example, this is the coherence of Ex-HxR, Ex-HyR, Ey-HxR, Ey-HyR, Hx-HxR, Hx-HyR, Hy-HxR, Hy-HyR

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        # now let's calculate coherency
        # abs(crosspower(A,B))^2/autopower(A)*autpower(B)
        for dataChan in self.specChans:
            for remoteChan in self.remoteChans:
                key = "{}{}RR".format(dataChan, remoteChan)
                for eIdx, eFreq in enumerate(self.evalFreq):
                    cohNom = np.power(
                        np.absolute(
                            self.getReferenceCrossPowerEval(dataChan, remoteChan, eIdx)
                        ),
                        2,
                    )
                    cohDenom = self.getAutoPowerEval(
                        dataChan, eIdx
                    ) * self.getRemoteAutoPowerEval(remoteChan, eIdx)
                    coh = cohNom / cohDenom
                    self.outData[eFreq][key] = coh
        return self.getOutData()

    def winRemoteEqnCoherence(self):
        """Calulates coherences for the remote reference solver equations

        For example, this is the coherence of Ex-HxR, Ex-HyR, Ey-HxR, Ey-HyR, Hx-HxR, Hx-HyR, Hy-HxR, Hy-HyR

        todo:
        Write more information in these comments

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        # now calculate out the relevant coherencies
        # here we calculate the coherency between <Ex,HyR> and <Hy,HyR> for example
        for iOut, outChan in enumerate(self.outChans):
            for iIn, inChan in enumerate(self.inChans):
                for iRemote, remoteChan in enumerate(self.remoteChans):
                    # calculate powers
                    c1c1 = smooth1d(
                        self.getReferenceCrossPower(outChan, remoteChan)
                        * np.conjugate(
                            self.getReferenceCrossPower(outChan, remoteChan)
                        ),
                        self.winLen,
                        self.winType,
                    )
                    c2c2 = smooth1d(
                        self.getReferenceCrossPower(inChan, remoteChan)
                        * np.conjugate(self.getReferenceCrossPower(inChan, remoteChan)),
                        self.winLen,
                        self.winType,
                    )
                    c1c2 = smooth1d(
                        self.getReferenceCrossPower(outChan, remoteChan)
                        * np.conjugate(self.getReferenceCrossPower(inChan, remoteChan)),
                        self.winLen,
                        self.winType,
                    )
                    # now interpolate
                    c1c1 = self.interpolateToEvalFreq(c1c1)
                    c2c2 = self.interpolateToEvalFreq(c2c2)
                    c1c2 = self.interpolateToEvalFreq(c1c2)
                    # now calculate the nominator and denominator
                    cohNom = np.power(np.absolute(c1c2), 2)
                    cohDenom = c1c1 * c2c2
                    coh = (
                        cohNom / cohDenom
                    )  # cast as float - discard complex part (complex part should be zero anyway)
                    # now need the coherencies for the evaluation frequencies
                    # this involves interpolation
                    key = "{}{}R-{}{}R".format(outChan, remoteChan, inChan, remoteChan)
                    for iFreq, eFreq in enumerate(self.evalFreq):
                        self.outData[eFreq][key] = coh[iFreq]
        return self.getOutData()

    def winRemoteAbsVal(self):
        """Absolute values of cross power spectral densities between remote refence channels and data channels

        This data can be useful for cross plotting

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        for eIdx, eFreq in enumerate(self.evalFreq):
            for iOut, outChan in enumerate(self.outChans):
                for iRemote, remoteChan in enumerate(self.remoteChans):
                    absOut = np.absolute(
                        self.getReferenceCrossPowerEval(outChan, remoteChan, eIdx)
                    )
                    keyOut = "abs{}{}R".format(outChan, remoteChan)
                    self.outData[eFreq][keyOut] = absOut
                    for iIn, inChan in enumerate(self.inChans):
                        absIn = np.absolute(
                            self.getReferenceCrossPowerEval(inChan, remoteChan, eIdx)
                        )
                        keyIn = "abs{}{}R".format(inChan, remoteChan)
                        self.outData[eFreq][keyIn] = absIn
        return self.getOutData()

    def winRemoteTransferFunction(self):
        """Calculate transfer function for the spectral data when remote reference is included too

        Returns
        -------
        Dict : 
            Dictionary with statistic values, indexed by [evaluation frequency][statistic component]
        """

        output = np.empty(
            shape=(self.evalFreq.size, self.outSize, self.inSize), dtype="complex"
        )
        for eIdx, eFreq in enumerate(self.evalFreq):
            # solve transfer function
            obs = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
            reg = np.empty(
                shape=(self.outSize, self.inSize, self.inSize), dtype="complex"
            )
            for i, outChan in enumerate(self.outChans):
                for j, remoteChan in enumerate(self.remoteChans):
                    # this is the observation row where,i is the observed output
                    # eIdx in the evaluation frequency
                    obs[i, j] = self.getReferenceCrossPowerEval(
                        outChan, remoteChan, eIdx
                    )
                    for k, inChan in enumerate(self.inChans):
                        reg[i, j, k] = self.getReferenceCrossPowerEval(
                            inChan, remoteChan, eIdx
                        )

            for i in range(0, self.outSize):
                observation = obs[i, :]
                predictors = reg[i, :, :]
                # now do the solution
                out, resids, squareResid, rank, s = olsModel(
                    predictors, observation, intercept=self.intercept
                )
                # out, resids, scale, weights	= mmestimateModel(predictors, observation, intercept=False)
                # not interested in the intercept (const) term
                if self.intercept:
                    output[eIdx, i] = out[1:]
                else:
                    output[eIdx, i] = out

            # calculate components of transfer function and res and phase
            for i in range(0, self.outSize):
                for j in range(0, self.inSize):
                    period = 1.0 / eFreq
                    res = 0.2 * period * np.power(np.absolute(output[eIdx, i, j]), 2)
                    phase = np.angle(output[eIdx, i, j], deg=True)
                    keyRes = self.outChans[i] + self.inChans[j] + "ResRR"
                    keyPhase = self.outChans[i] + self.inChans[j] + "PhaseRR"
                    self.outData[eFreq][keyRes] = res
                    self.outData[eFreq][keyPhase] = phase
                    # add the components
                    keyReal = self.outChans[i] + self.inChans[j] + "RealRR"
                    keyImag = self.outChans[i] + self.inChans[j] + "ImagRR"
                    self.outData[eFreq][keyReal] = output[eIdx, i, j].real
                    self.outData[eFreq][keyImag] = output[eIdx, i, j].imag
        # set transfer function calculated as true
        # saves having to do it again
        self.remoteCalculated = True
        return self.getOutData()

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Default options")
        textLst.append("\tInput Chans = {}".format(listToString(self.getInChans())))
        textLst.append("\tOutput Chans = {}".format(listToString(self.getOutChans())))
        textLst.append(
            "\tRemote Chans = {}".format(listToString(self.getRemoteChans()))
        )
        textLst.append("\tPowers = {}".format(listToString(self.getPSDChans())))
        textLst.append(
            "\tCoherence pairs = {}".format(listToString(self.getCohPairs()))
        )
        textLst.append(
            "\tPartial coherence = {}".format(listToString(self.getPolDirs()))
        )
        if len(self.getEvalFreq()) == 0:
            textLst.append("Evaluation frequencies = {}")
        else:
            textLst.append(
                "Evaluation frequencies = {}".format(arrayToString(self.getEvalFreq()))
            )
        return textLst



import numpy as np
from typing import List

# import from package
from resistics.calculators.calculator import Calculator
from resistics.calculators.decimationParameters import DecimationParams
from resistics.utilities.utilsConfig import loadConfig


class WindowParams(Calculator):
    """WindowParams

    Calculates window sizes for each decimation level based on minimum allowable window size and overlap. Alternatively, users can directly set windowSizes and overlaps. 

    Attributes
    ----------
    decParams : float
        Exposure in seconds.
    minSize : int
        Minimum allowable window size
    minOlap : int
        Minimum allowable overlap
    windowFactor : float
        Window size calculated by sampling frequency / windowFactor to ensure good frequency domain resolution
    overlapFraction : float
        Overlap size as a fraction of window size
    windows : List[int], np.ndarray
        List or array with window sizes for each decimation level
    overlaps : List[int], np.ndarray
        List or array with overlap sizes for each decimation level

    Methods
    -------
    __init__(decParams)
        Initialise with information about the decimation parameters
    getWindowSize(iDec)
        Returns window size for decimation level iDec (starts at 0)
    getOverlap(iDec)
        Returns overlap size for decimation level iDec (starts at 0)
    setWindowParameters(windowSizes, windowOverlaps)
        Directly set window sizes and overlaps
    setMinParams(minSize, minOverlap)
        Set minimum allowable window size and overlap which will be honoured when window sizes by decimation level are automatically calculated
    calcParameters(windowFactor, overlapFraction)
        Calculate window and overlap sizes for each decimation level
    printList()
        Class status returned as list of strings          
    """

    def __init__(self, decParams: DecimationParams) -> None:
        """Initialise with decParams and default window parameters

        Parameters
        ----------
        decParams : int
            Decimation level
        """

        self.decParams = decParams
        config = loadConfig()
        self.minSize = config["Window"]["minwindowsize"]
        self.minOlap = config["Window"]["minoverlapsize"]
        self.windowFactor = config["Window"]["windowfactor"]
        self.overlapFraction = config["Window"]["overlapfraction"]
        self.calcParameters(self.windowFactor, self.overlapFraction)

    def getWindowSize(self, decLevel: int) -> int:
        """Get window size for decimation level

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        windowSize : int
            Window size for decimation level
        """

        return self.windows[decLevel]

    def getOverlap(self, decLevel: int) -> int:
        """Get window overlap for decimation level

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        windowOverlap : int
            Window overlap for decimation level
        """

        return self.overlaps[decLevel]

    def setWindowParameters(
        self, windowSizes: List[int], windowOverlaps: List[int]
    ) -> None:
        """Directly set window parameters rather than calculating them out

        If a user is not satisfied with the window size calculation, window parameters can be defined explicitly here

        Parameters
        ----------
        windowSizes : List[int]
            List of window sizes for each decimation level
        windowOverlaps : List[int]
            List of window overlaps for each decimation level
        """

        if (
            len(windowSizes) != self.decParams.numLevels
            or len(windowOverlaps) != self.decParams.numLevels
        ):
            print(
                "Error: not enough window sizes given. Must be equal to number of decimation levels"
            )
            return
        self.windows = windowSizes
        self.overlaps = windowOverlaps

    def setMinParams(self, minSize: int, minOlap: int) -> None:
        """Change default minimum window size and overlap parameters

        These values will be honoured when automatically calculating window sizes in calcParameters

        Parameters
        ----------
        minSize : int
            List of window sizes for each decimation level
        minOlap : int
            List of window overlaps for each decimation level
        """

        self.minSize = minSize
        self.minOlap = minOlap
        self.calcParameters(self.windowFactor, self.overlapFraction)

    def calcParameters(self, windowFactor, overlapFraction) -> None:
        """Calculate window size and overlap for each decimation level based on minimum allowable parameters (minSize, minOlap)

        The window and overlap sizes (number of samples) are calculated based on the following:

        Window size = frequency at decimation level / windowFactor
        Overlap size = Window size * overlapFraction

        The window size is calculated based on the sampling frequency of the decimation level to ensure good frequency domain resolution


        Parameters
        ----------
        windowFactor : float
            Window size is calculated as frequency at decimation level / windowFactor
        overlapFraction : float
            Overlap size as a fraction of the window size
        """

        self.windows = np.ones(shape=(self.decParams.numLevels), dtype=int)
        self.overlaps = np.ones(shape=(self.decParams.numLevels), dtype=int)
        decFreq = self.decParams.decFrequencies
        for il in range(0, self.decParams.numLevels):
            self.windows[il] = int(decFreq[il] / windowFactor)
            if self.windows[il] < self.minSize:
                self.windows[il] = self.minSize
            self.overlaps[il] = int(self.windows[il] * overlapFraction)
            if self.overlaps[il] < self.minOlap:
                self.overlaps[il] = self.minOlap

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append(
            "Number of decimation levels = {:d}".format(self.decParams.numLevels)
        )
        decFrequencies = self.decParams.decFrequencies
        for il in range(0, self.decParams.numLevels):
            textLst.append(
                "Level = {:d}, sample freq. [Hz] = {:.6f}, sample rate [s] = {:6f}".format(
                    il, decFrequencies[il], 1.0 / decFrequencies[il]
                )
            )
            textLst.append(
                "Window size = {:d}, window duration [s] = {:f}".format(
                    self.windows[il], (self.windows[il] - 1) / decFrequencies[il]
                )
            )
            textLst.append(
                "Window overlap = {:d}, overlap duration [s] = {:f}".format(
                    self.overlaps[il], (self.overlaps[il] - 1) / decFrequencies[il]
                )
            )
        return textLst



import numpy as np
import scipy.signal as signal
from datetime import timedelta
from typing import List

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.timeData import TimeData
from resistics.calculators.decimationParameters import DecimationParams
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsMath import intdiv
from resistics.utilities.utilsPrint import (
    generalPrint,
    warningPrint,
    blockPrint,
    arrayToStringInt,
)
from resistics.utilities.utilsFilter import downsampleData


class Decimator(Calculator):
    """Decimate time data

    Decimates time data by factors until the minimum number of required samples is reached. When a downsample factor is too large, downsampling is performed in multiple steps to maintain accuracy of result.

    Attributes
    ----------
    timeData : TimeData
        timeData object to decimate
    sampleFreq : float
        Sampling frequency of time data in Hz
    chans : List[str]
        Channels in time data
    numSamples : int
        Number of samples in timeData
    decParams : DecimationParams
        A DecimationParams object holding decimation information
    minSamples : int
        Minimum required samples to decimate
    level : int
        Current decimation level
    maxDownSampleFactor : int
        Max allowable downsampling in one go. Downsampling becomes less accurate at large downsample factors

    Methods
    -------
    __init__(timeData, decParams)
        Initialise Decimator with a TimeData object and DecimationParams object
    incrementLevel()
        Downsample the timeData to the next decimation level
    downsample(downsampleFactor)
        Do the downsampling
    printList()
        Class status returned as list of strings  
    """

    def __init__(self, timeData: TimeData, decParams: DecimationParams) -> None:
        """Initialise with timeData and decimation parameters
    
        Parameters
        ----------
        timeData : TimeData
            The time data to decimate
        decParams : DecimationParams
            Decimation parameters for performing the decimation
        """

        self.timeData: TimeData = timeData
        self.sampleFreq: float = timeData.sampleFreq * 1.0
        self.chans: List = timeData.chans
        self.numSamples: int = timeData.numSamples
        self.decParams: DecimationParams = decParams
        config = loadConfig()
        self.minSamples: int = config["Decimation"]["minsamples"]
        self.level: int = -1
        self.maxDownsampleFactor: int = 8

    def incrementLevel(self) -> bool:
        """Downsample to the next decimation level

        Returns
        -------
        out : bool
            True if downsampling completed successfully. False otherwise

        Notes
        -----
        When the downsampling factor is too large, downsampling is performed in multiple steps. Downsampling will become increasingly inaccurate using the scipy routine when factor is too large        
        """

        # increment level, 0 is the first level
        self.level = self.level + 1
        downsampleFactor = self.decParams.getIncrementalFactor(self.level)

        # if downsample factor is greater than maxDownsampleFactor, downsample in multiple steps
        numDownsamples = 1
        downsampleList = [downsampleFactor]
        if downsampleFactor > self.maxDownsampleFactor:
            # this should give an integer
            numDownsamples = intdiv(downsampleFactor, self.maxDownsampleFactor)
            downsampleList = [self.maxDownsampleFactor, numDownsamples]
            # print info
            self.printText(
                "Downsample factor of {:d} greater than max decimation factor {:d}.".format(
                    downsampleFactor, self.maxDownsampleFactor
                )
            )
            self.printText(
                "Downsampling in multiple decimations given by factors: {}".format(
                    arrayToStringInt(downsampleList)
                )
            )

        for iDS in range(0, numDownsamples):
            check = self.downsample(downsampleList[iDS])
            if not check:  # check outcome of decimation
                return False
            # otherwise, everything ok, update class vars and return True
            self.sampleFreq = self.timeData.sampleFreq
            self.numSamples = self.timeData.numSamples
        return True

    def downsample(self, downsampleFactor: int) -> bool:
        """Downsample time data

        Parameters
        ----------
        downsampleFactor : int
            Downsampling factor        

        Returns
        -------
        out : bool
            True if downsampling completed successfully. False otherwise

        Notes
        -----
        When the downsampling causes number of simples to fall below minSamples, downsampling is not performed. The function returns False in this situation      
        """

        # check to see not at max level
        if self.level >= self.decParams.numLevels:
            self.printWarning(
                "Error, number of decimation levels exceeded, returning no data"
            )
            return False
        # if downsample factor is 1, nothing to do
        if downsampleFactor == 1:
            return True

        # downsampling reduces the number of samples by downsample factor
        # if new number of samples is too small, return False
        if self.numSamples / downsampleFactor < self.minSamples:
            self.printWarning(
                "Next decimation level has less than {} samples. Decimation is exiting.\nSet minimum of samples required using decimator.setMinSamples().".format(
                    self.minSamples
                )
            )
            return False

        # do the resampling
        self.timeData.data = downsampleData(self.timeData.data, downsampleFactor)
        # update the rest of the timeData object
        self.timeData.sampleFreq = self.timeData.sampleFreq / downsampleFactor
        self.timeData.numSamples = self.timeData.data[self.chans[0]].size
        # start time stays the same, but end time needs to change
        self.timeData.stopTime = self.timeData.startTime + timedelta(
            seconds=(1.0 / self.timeData.sampleFreq) * (self.timeData.numSamples - 1)
        )
        self.timeData.addComment(
            "Time data decimated from {} Hz to {} Hz, new start time {}, new end time {}".format(
                self.sampleFreq,
                self.timeData.sampleFreq,
                self.timeData.startTime,
                self.timeData.stopTime,
            )
        )
        return True

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Current level = {:d}".format(self.level))
        if self.level == -1:
            textLst.append("This is the initial level - no decimation has occured")
        textLst.append("Current sample freq. [Hz] = {:.6f}".format(self.sampleFreq))
        textLst.append("Current sample rate [s] = {:.6f}".format(1.0 / self.sampleFreq))
        textLst.append("Current number of samples = {:d}".format(self.numSamples))
        return textLst



import numpy as np
import math
from typing import Dict


def removeZeros(data: Dict):
    """Remove a stretch of zeros in the data

    This function finds a stretch of zeros and tries to fill them in with better data i.e. interpolated data or some such.

    Parameters
    ----------
    data : Dict
        Dictionary of data with channel as key and a np.ndarray as value

    Returns
    -------
    Dict
        Dictionary of data with channel as key and a np.ndarray as value (with zero stretches removed)
    """

    for chan in data:
        data[chan] = removeZerosSingle(data[chan])
    return data


def removeZerosSingle(data: np.ndarray) -> np.ndarray:
    """Remove a stretch of zeros in a data array

    This function finds a stretch of zeros and tries to fill them in with better data i.e. interpolated data or some such.

    Parameters
    ----------
    data : np.ndarray
        Array of data

    Returns
    -------
    np.ndarray
        Array of data with zeros removed
    """

    eps = 0.000000001  # use this because of floating point precision
    # set an x array
    x = np.arange(data.size)
    # find zero locations - this returns a tuple, take the first index
    zeroLocs = np.where(np.absolute(data) < eps)[0]
    if len(zeroLocs) == 0:
        return data  # no zeros to remove

    # now want to find consecutive zeros
    grouped = groupConsecutive(zeroLocs)
    indicesToFix = []
    # now find groups of 3+
    for g in grouped:
        if g.size >= 20:
            indicesToFix = indicesToFix + list(g)
    # now have the indices we want to fix
    # can go about interpolating values there
    indicesToFix = np.array(sorted(indicesToFix))
    mask = np.ones(data.size, np.bool)
    mask[indicesToFix] = 0
    data[indicesToFix] = np.interp(indicesToFix, x[mask], data[mask])
    return data


def groupConsecutive(vals: np.ndarray, stepsize: int = 1):
    """Takes an array of values and splits it into consecutive sections of stepsize

    In general, the stepsize is 1.
    
    Parameters
    ----------
    vals : np.ndarray
        A set of values to split into consecutive sections
    stepsize : int
        The stepsize between values that means they are consecutive

    Examples
    --------
    An array of [1,2,3,5,6,7,10,12,13] would be split into consecutive sections [1,2,3], [5,6,7], [10], [12,13]
    """

    return np.split(vals, np.where(np.diff(vals) != stepsize)[0] + 1)


def removeNans(data: Dict):
    """Remove NaNs in the data

    This function finds NaNs in the data and tries to fill them in with better data i.e. interpolated data or some such.

    Parameters
    ----------
    data : Dict
        Dictionary of data with channel as key and a np.ndarray as value

    Returns
    -------
    Dict
        Dictionary of data with channel as key and a np.ndarray as value (with zero stretches removed)
    """

    # find nan in the dataset and removes the values
    for chan in data:
        data[chan] = removeNansSingle(data[chan])
    return data


def removeNansSingle(data):
    """Remove NaNs in a data array

    This function finds NaNs in the np.ndarray and tries to fill them in with better data i.e. interpolated data or some such.

    Parameters
    ----------
    data : np.ndarray
        Array of data

    Returns
    -------
    np.ndarray
        Array of data with zeros removed
    """

    # set an x array
    x = np.arange(data.size)
    # find locations of nans - this is a bool array with True in locations with nan values
    nanLocs = np.isnan(data)
    # if no nans, do nothing
    if not np.any(nanLocs):
        return data  # no nans to remove

    # create mask
    mask = np.ones(data.size, np.bool)
    mask[nanLocs] = 0  # using numpy indexing with bool arrays
    # no need to group, want to remove every nan
    data[nanLocs] = np.interp(x[nanLocs], x[mask], data[mask])
    return data



from datetime import datetime
import numpy as np
from typing import Union, List, Set


def generalPrint(pre: str, info: str) -> None:
    """Print to terminal

    Parameters
    ----------
    pre : str
        String describing where the information is being output from 
    info : str
        Information string
    """

    print("{} {}: {}".format(datetime.now().strftime("%H:%M:%S"), pre, info))


def breakPrint() -> None:
    """Prints a break in the termainal to make things easier to read"""

    print("---------------------------------------------------")
    print("")
    print("---------------------------------------------------")


def blockPrint(pre: str, text: Union[List, str]) -> None:
    """Prints a block of information to the terminal with separators

    Parameters
    ----------
    pre : str
        String describing where the information is being output from 
    text : str, List
        Information string or a list of strings to print out in the black
    """

    generalPrint(pre, "####################")
    generalPrint(pre, "{} INFO BEGIN".format(pre.upper()))
    generalPrint(pre, "####################")
    if isinstance(text, str):
        generalPrint(pre, text)
    else:
        for t in text:
            generalPrint(pre, t)
    generalPrint(pre, "####################")
    generalPrint(pre, "{} INFO END".format(pre.upper()))
    generalPrint(pre, "####################")


def warningPrint(pre: str, info: str) -> None:
    """Print a warning to the terminal

    Parameters
    ----------
    pre : str
        String describing where the information is being output from 
    info : str
        Information string
    """

    print("{} {}: {}".format(datetime.now().strftime("%H:%M:%S"), pre.upper(), info))


def errorPrint(pre: str, info: str, quitRun: bool = False):
    """Print an error

    Parameters
    ----------
    pre : str
        String describing where the information is being output from 
    info : str
        Information string
    quitRun : bool, optional
        Bool flag for quitting execution of the code due to the error
    """

    print(
        "ERROR: {} {}: {}".format(
            datetime.now().strftime("%H:%M:%S"), pre.upper(), info
        )
    )
    if quitRun:
        print("Exiting...")
        exit()


def breakComment() -> str:
    """Returns a string to show a break in data comments
    
    Returns
    -------
    str
        A separator string
    """

    return "---------------------------------------------------"


def arrayToString(data: np.ndarray, tabs: bool = False, decimals: int = 8) -> str:
    """Convert an array to a string separated by commas

    Parameters
    ----------
    data : np.ndarray
        Data array
    tabs : bool, optional
        Bool flag for using tabs instead of commas 
    decimals : int, optional
        Number of decimal places to print

    Returns
    -------
    out : str
        Output string
    """

    outputStr: str = ""
    sep: str = ", "
    if tabs:
        sep = "\t"

    for d in data:
        outputStr = outputStr + "{num:.{dec}f}{sep}".format(
            num=d, dec=decimals, sep=sep
        )
    outputStr = outputStr.strip()
    outputStr = outputStr.rstrip(",")
    return outputStr


def arrayToStringSci(data: np.ndarray) -> str:
    """Convert an array to a string of scientific numbers separated by commas

    Parameters
    ----------
    data : np.ndarray
        Data array

    Returns
    -------
    out : str
        Output string
    """

    outputStr = ""
    for d in data:
        outputStr = outputStr + "{:.6e}, ".format(d)
    outputStr = outputStr.strip()
    return outputStr.rstrip(",")


def arrayToStringInt(data: np.ndarray) -> str:
    """Convert an array to a string of integers numbers separated by commas

    Parameters
    ----------
    data : np.ndarray
        Data array

    Returns
    -------
    out : str
        Output string
    """

    outputStr = ""
    for d in data:
        if isinstance(d, float):
            d = int(d)
        outputStr = outputStr + "{:d}, ".format(d)
    outputStr = outputStr.strip()
    return outputStr.rstrip(",")


def listToString(lst: List, floatlim: bool = True) -> str:
    """Convert a list to a comma separated string

    Parameters
    ----------
    lst : list
        List

    Returns
    -------
    out : str
        Output string
    """

    outputStr = ""
    for val in lst:
        outputStr = outputStr + "{}, ".format(val)
    outputStr = outputStr.strip()
    return outputStr.rstrip(",")


def list2rangesFormatter(
    start: Union[int, str], end: Union[int, str], step: Union[int, str]
) -> str:
    """A string for showing a range

    Parameters
    ----------
    start : int
        Start point
    end : int
        End point
    step : int
        Step

    Returns
    -------
    out : str
        Output string
    """

    return "{}-{}:{}".format(start, end, step)


def list2ranges(data: Union[List, Set]):
    """Convert a list of numbers to a list of ranges

    Parameters
    ----------
    data : list, set
        List or set of numbers

    Returns
    -------
    out : str
        Output string
    """

    lst = data
    if type(data) is set:
        lst = list(data)
    lst = sorted(lst)
    n = len(lst)
    result = []
    resultVals = []
    scan = 0
    while n - scan > 2:
        step = lst[scan + 1] - lst[scan]
        if lst[scan + 2] - lst[scan + 1] != step:
            result.append(str(lst[scan]))
            resultVals.append([lst[scan], lst[scan], 0])
            scan += 1
            continue

        for j in range(scan + 2, n - 1):
            if lst[j + 1] - lst[j] != step:
                result.append(list2rangesFormatter(lst[scan], lst[j], step))
                resultVals.append([lst[scan], lst[j], step])
                scan = j + 1
                break
        else:
            result.append(list2rangesFormatter(lst[scan], lst[-1], step))
            resultVals.append([lst[scan], lst[-1], step])
            return ",".join(result)

    if n - scan == 1:
        result.append(str(lst[scan]))
        resultVals.append([lst[scan], lst[scan], 0])
    elif n - scan == 2:
        result.append(",".join(map(str, lst[scan:])))
        resultVals.append([lst[scan], lst[scan], 0])
        resultVals.append([lst[scan + 1], lst[scan + 1], 0])

    return ",".join(result)



import os
from copy import deepcopy
from datetime import datetime, timedelta
import numpy as np
from typing import List, Dict, Any

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.spectrumData import SpectrumData
from resistics.utilities.utilsWindow import gIndex2datetime
from resistics.utilities.utilsPrint import breakComment


class SpectrumReader(IOHandler):
    """Reads in spectra data for measurements

    Spectrum reader reads in the info file for the spectra data and .dat files (for ascii formatted spectra data) and .bin files (for binary formatted data).

    Spectra files are located in:
    project -> specData -> site -> datapath -> spectra data files

	Attributes
	----------
    datapath : str 
        Data root for spectra data
    headerKeys :
        Spectra file header keys
    headers : Dict
        Spectra file header values
    comments: List[str]
        Spectra file comments
    dataType :
        Data type of spectra data
    dataByteSize : int 
        Byte size of a single data point
    filepath : str
        Filepath for spectra files
    file : bool 
        The data file 

	Methods
	-------
	__init__(datapath)
		Initialise the SpectrumReader
	getReferenceTime()
		Get the reference time used for the spectrum calculation
    def getChannels()
        Get the channels in the spectra data
    getNumChannels()
        Get the number of channels
    getSampleFreq()
        Get the sampling frequency of the data
    getNumWindows()
        Get the number of windows in the spectra file        
    getWindowSize()
        Get the window size used for the data
    getWindowOverlap()
        Get the window overlap size
    def getDataSize()
        Get the number of samples in a spectra window
    getGlobalOffset()
        Get the window global offset
    getGlobalRange()
        Get the window global indices range
    getFrequencyArray()
        Get the frequency array of the frequencies of the spectra data points
    getComments()
        Get a deepcopy of the comments
    openBinaryForReading(filename, fileInc):
        Open a binary spectra file given by filename and fileInc (the decimation level)
    readBinaryWindowLocal(localIndex)
        Read a spectra window given by a local index in binary format
    readBinaryWindowGlobal(globalIndex)
        Read a spectra window given by a global index (relative to reference time) in binary format
    openAsciiForReading(filename, fileInc):
        Open a ascii spectra file given by filename and fileInc (the decimation level)
    readAsciiWindowLocal(localIndex)
        Read a spectra window given by a local index in ascii format
    readAsciiWindowGlobal(globalIndex)
        Read a spectra window given by a global index (relative to reference time) in ascii format
    readInfoFile(filepath)
        Read the spectra information file given by filepath
    getInfoValue(header, line)
        Put header value in the correct format for header and line
    readCommentsFile(filepath):
        Read the comments file given by filepath   
    getFileBase(filename, fileInc):
        Get the spectra file name
    closeFile()
        Close spectra data file 
    printList()
        Class status returned as list of strings       
	"""

    def __init__(self, datapath: str) -> None:
        """Initialise

        Parameters
        ----------
        datapath : str, optional
            Path to spectra directory
        """

        self.datapath: str = datapath
        if not os.path.exists(self.datapath):
            self.printWarning("Directory {} does not exist".format(self.datapath))
        self.headerKeys: List[str] = [
            "Reference time",
            "Sample frequency",
            "Window size",
            "Window overlap",
            "Global offset",
            "Number of windows",
            "Number of channels",
            "Data size",
            "Channels",
        ]
        self.headers: Dict = {}
        self.comments: List[str] = []
        self.dataType = np.dtype("complex64")
        self.dataByteSize: int = self.dataType.itemsize
        self.filepath: str = ""
        self.file: bool = False

    # def getDataRoot(self):
    #     return self.datapath

    # def getHeaders(self):
    #     return self.headerKeys

    # def getHeaderDict(self):
    #     return self.headers

    # def getComments(self):
    #     return self.comments

    # def setDataRoot(self, datapath):
    #     self.datapath = datapath

    def getReferenceTime(self) -> datetime:
        """Get reference time for spectra calculation

        Returns
        -------
        datetime
            The reference time used for the spectra calculation
        """

        return self.headers["Reference time"]

    def getSampleFreq(self) -> float:
        """Get sampling frequency of time data used for spectra calculation

        Returns
        -------
        float
            Sampling frequency of the time data
        """

        return float(self.headers["Sample frequency"])

    def getNumChannels(self) -> int:
        """Get the number of channels

        Returns
        -------
        int
            The number of channels
        """

        return int(self.headers["Number of channels"])

    def getChannels(self) -> List[str]:
        """Get the channels in the spectra data

        Returns
        -------
        List[str]
            List of channels
        """

        return self.headers["Channels"]

    def getNumWindows(self) -> int:
        """Get number of windows for which spectra have been calculated

        Returns
        -------
        int
            The number of windows
        """

        return int(self.headers["Number of windows"])

    def getWindowSize(self) -> int:
        """Get the size of the time data window in samples

        Returns
        -------
        int
            Size of time data window in samples
        """

        return int(self.headers["Window size"])

    def getWindowOverlap(self) -> int:
        """Get the size of the time data overlap in samples

        Returns
        -------
        int
            Size of time data overlap in samples
        """

        return int(self.headers["Window overlap"])

    def getDataSize(self) -> int:
        """Get the size of the corresponding spectrum data for a window

        Returns
        -------
        int
            Number of samples in one spectrum window
        """

        return int(self.headers["Data size"])

    def getGlobalOffset(self) -> int:
        """Get global window offset

        The global window offset references windows to the reference time rather than the start time of the time data

        Returns
        -------
        int
            Global window offset
        """

        return int(self.headers["Global offset"])

    def getGlobalRange(self) -> List[int]:
        """Get the range of window global indices

        The first global index is given by getGlobalOffset() as the counting starts from zero. The last global index is self.getGlobalOffset() + self.getNumWindows() - 1.

        Returns
        -------
        List[int]
            List with two elements, the first being the global index of the first spectrum window and the second, the global index of the last spectrum window
        """

        return [
            self.getGlobalOffset(),
            self.getGlobalOffset() + self.getNumWindows() - 1,
        ]

    def getFrequencyArray(self) -> np.ndarray:
        """Returns the frequency array

        Frequency array are the frequency points in the spectra

        Returns
        -------
        np.ndarray
            Frequency array
        """

        return np.linspace(0, self.getSampleFreq() / 2.0, self.getDataSize())

    def getComments(self) -> List[str]:
        """Get a deepcopy of the comments
        
        Returns
        -------
        List[str]
            Dataset comments as a list of strings
        """

        return deepcopy(self.comments)

    def openBinaryForReading(self, filename: str, fileInc: int) -> bool:
        """Open a binary data file for reading

        Parameters
        ----------
        filename: str
            Filename of spectra files
        fileInc : int
            The decimation level
        """

        filebase = self.getFileBase(filename, fileInc)
        filepathInfo = os.path.join(self.datapath, filebase + ".info")
        filepathComments = os.path.join(self.datapath, "comments.txt")
        self.filepath = os.path.join(self.datapath, filebase + ".bin")
        # check files exist
        if not os.path.exists(filepathInfo) or not os.path.exists(self.filepath):
            self.printWarning(
                "No data found in either {} or {}".format(filepathInfo, self.filepath)
            )
            return False
        # read info file
        self.readInfoFile(filepathInfo)
        self.readCommentsFile(filepathComments)
        self.channelByteSize = self.dataByteSize * self.getDataSize()
        self.windowByteSize = self.channelByteSize * self.getNumChannels()
        # set file to filepath - this is because binary files do not require opening using memap
        # self.file = self.filepath
        return True

    def readBinaryWindowLocal(self, localIndex: int) -> SpectrumData:
        """Get spectrum data for a window defined by a local index (for binary formatted data)

        Parameters
        ----------
        localIndex: int
            The local index
        """

        if localIndex >= self.getNumWindows():
            self.printWarning("Local index {:d} out of bounds".format(localIndex))
            self.printWarning(
                "Min index = {:d}, Max index = {:d}".format(0, self.getNumWindows() - 1)
            )
        # with binary files, want the correct bytes
        byteOff = localIndex * self.windowByteSize
        data = {}
        for cI, c in enumerate(self.getChannels()):
            chanOff = cI * self.channelByteSize
            data[c] = np.memmap(
                self.filepath,
                dtype=self.dataType,
                mode="r",
                offset=byteOff + chanOff,
                shape=(self.getDataSize()),
            )
        # return data
        startTime, stopTime = gIndex2datetime(
            localIndex + self.getGlobalOffset(),
            self.getReferenceTime(),
            self.getSampleFreq(),
            self.getWindowSize(),
            self.getWindowOverlap(),
        )
        return SpectrumData(
            windowSize=self.getWindowSize(),
            dataSize=self.getDataSize(),
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=self.comments,
        )

    def readBinaryWindowGlobal(self, globalIndex: int) -> SpectrumData:
        """Get spectrum data for a window defined by a global index (for binary formatted data)

        Parameters
        ----------
        globalIndex: int
            The global index
        """

        if (
            globalIndex >= self.getNumWindows() + self.getGlobalOffset()
            or globalIndex < self.getGlobalOffset()
        ):
            self.printWarning("Global index {:d} out of bounds".format(globalIndex))
            self.printWarning(
                "Min index = {:d}, Max index = {:d}".format(
                    self.getGlobalOffset(),
                    self.getGlobalOffset() + self.getNumWindows() - 1,
                )
            )
        # convert global index to local index and return readAsciiWindowLocal
        localIndex = globalIndex - self.getGlobalOffset()
        return self.readBinaryWindowLocal(localIndex)

    def openAsciiForReading(self, filename: str, fileInc: int) -> bool:
        """Open a ascii data file for reading

        Parameters
        ----------
        filename: str
            Filename of spectra files
        fileInc : int
            The decimation level
        """

        filebase = self.getFileBase(filename, fileInc)
        filepathInfo = os.path.join(self.datapath, filebase + ".info")
        self.filepath = os.path.join(self.datapath, filebase + ".dat")
        # check files exist
        if not os.path.exists(filepathInfo) or not os.path.exists(self.filepath):
            self.printWarning(
                "No data found in either {} or {}".format(filepathInfo, self.filepath)
            )
            return False
        # read info file
        self.readInfoFile(filepathInfo)
        # open file for reading
        self.file = open(self.filepath, "rb")
        # run through and find line endings
        self.lineOffset = []
        offset = 0
        for line in self.file:
            self.lineOffset.append(offset)
            offset += len(line)
        self.file.seek(0)
        return True

    def readAsciiWindowLocal(self, localIndex: int) -> SpectrumData:
        """Get spectrum data for a window defined by a local index (for ascii formatted data)

        Parameters
        ----------
        localIndex: int
            The local index
        """

        if localIndex >= self.getNumWindows():
            self.printWarning("Local index {:d} out of bounds".format(localIndex))
            self.printWarning(
                "Min index = {:d}, Max index = {:d}".format(0, self.getNumWindows() - 1)
            )
        # with ascii files, want the correct lines
        # find line where local index starts
        windowStartLine = localIndex * self.getNumChannels()
        data = {}
        for cI, c in enumerate(self.getChannels()):
            indexC = windowStartLine + cI
            self.file.seek(self.lineOffset[indexC])
            line = self.file.readline()
            data[c] = np.loadtxt(line.strip().split(","), dtype=complex)
        # return data
        startTime, stopTime = gIndex2datetime(
            localIndex + self.getGlobalOffset(),
            self.getReferenceTime(),
            self.getSampleFreq(),
            self.getWindowSize(),
            self.getWindowOverlap(),
        )
        return SpectrumData(
            windowSize=self.getWindowSize(),
            dataSize=self.getDataSize(),
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=self.comments,
        )

    def readAsciiWindowGlobal(self, globalIndex: int) -> SpectrumData:
        """Get spectrum data for a window defined by a global index (for ascii formatted data)

        Parameters
        ----------
        globalIndex: int
            The global index
        """

        if (
            globalIndex >= self.getNumWindows() + self.getGlobalOffset()
            or globalIndex < self.getGlobalOffset()
        ):
            self.printWarning("Global index {:d} out of bounds".format(globalIndex))
            self.printWarning(
                "Min index = {:d}, Max index = {:d}".format(
                    self.getGlobalOffset(),
                    self.getGlobalOffset() + self.getNumWindows() - 1,
                )
            )
        # convert global index to local index and return readAsciiWindowLocal
        localIndex = globalIndex - self.getGlobalOffset()
        return self.readAsciiWindowLocal(localIndex)

    def readInfoFile(self, filepath: str) -> None:
        """Read the info file for the spectra

        Parameters
        ----------
        filepath : str
            Info file path
        """

        infoFile = open(filepath, "r")
        lines = infoFile.readlines()
        infoFile.close()
        # loop through all headers and get values
        for h in self.headerKeys:
            for l in lines:
                if h in l:
                    self.headers[h] = self.getInfoValue(h, l)
                    break

    def getInfoValue(self, header: str, line: str) -> Any:
        """Format some of the info file values

        Parameters
        ----------
        header : str
            The header 
        line : str
            The line from the info file
        """

        split = line.split("=")
        split[1] = split[1].strip()
        if header == "Channels":
            return split[1].split()
        elif header == "Reference time":
            return datetime.strptime(split[1], "%Y-%m-%d %H:%M:%S.%f")
        elif header == "Sample frequency":
            return float(split[1])
        else:
            return int(float(split[1]))

    def readCommentsFile(self, filepath: str) -> None:
        """Read comments file from filepath

        Parameters
        ----------
        filepath : str
            Comments file path
        """

        if os.path.exists(filepath):
            f = open(filepath, "r")
            self.comments = f.readlines()
            f.close()
            for idx, comment in enumerate(self.comments):
                self.comments[idx] = comment.rstrip()
        # add a new comment
        self.comments.append("Reading spectra data in path {}".format(self.datapath))

    def getFileBase(self, filename: str, fileInc: int) -> str:
        """Read comments file from filepath

        Parameters
        ----------
        filename: str
            Filename of spectra files
        fileInc : int
            The decimation level
        """

        return filename + "{:02d}".format(fileInc)

    def closeFile(self):
        """Close spectra file"""

        if self.filepath != "" and self.file:
            self.printText("Closing file {}".format(self.filepath))
            self.file.close()
            self.filepath = ""
        else:
            print("No file open")

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Data root = {}".format(self.datapath))
        if len(self.headers) > 0:
            textLst.append("Filepath = {}".format(self.filepath))
            for h in self.headerKeys:
                textLst.append("{} = {}".format(h, self.headers[h]))
            if len(self.comments) > 0:
                textLst.append("Comments")
                for comment in self.comments:
                    textLst.append("\t{}".format(comment))
            else:
                textLst.append("No comments")
        return textLst




import sys
import os
import numpy as np
from datetime import datetime
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import LogNorm
from typing import List, Union, Dict

# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.siteData import SiteData
from resistics.dataObjects.spectrumData import SpectrumData
from resistics.calculators.calibrator import Calibrator
from resistics.calculators.decimator import Decimator
from resistics.calculators.windower import Windower
from resistics.calculators.spectrumCalculator import SpectrumCalculator
from resistics.ioHandlers.spectrumReader import SpectrumReader
from resistics.ioHandlers.spectrumWriter import SpectrumWriter
from resistics.utilities.utilsChecks import parseKeywords, isMagnetic
from resistics.utilities.utilsPrint import listToString, breakComment
from resistics.utilities.utilsPlotter import (
    savePlot,
    colorbar2dSpectra,
    colorbarMultiline,
    plotOptionsSpec,
)
from resistics.utilities.utilsProject import (
    projectText,
    projectError,
    getCalibrator,
    getDecimationParameters,
    getWindowParameters,
    applyPolarisationReversalOptions,
    applyCalibrationOptions,
    applyFilterOptions,
    applyNotchOptions,
)


def getSpecReader(
    projData: ProjectData, site: str, meas: str, **kwargs
) -> SpectrumReader:
    """Get the spectrum reader for a measurement

    Parameters
    ----------
    site : str
        Site for which to get the spectra reader
    meas : str
        The measurement
    options : Dict
        Options in a dictionary
    declevel : int, optional
        Decimation level for which to get data
    specdir : str, optional
        String that specifies spectra directory for the measurement

    Returns
    -------
    SpectrumReader
        The spectrum reader object
    """

    options = {}
    options["declevel"]: int = 0
    options["specdir"]: str = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    siteData = projData.getSiteData(site)
    measurements = siteData.getMeasurements()
    if meas not in measurements:
        projectError("Measurement directory {} not found".format(meas), quitRun=True)

    # create the spectrum reader
    specReader = SpectrumReader(
        os.path.join(siteData.getMeasurementSpecPath(meas), options["specdir"])
    )
    specReader.printInfo()

    # open the spectra file for the current decimation level
    check = specReader.openBinaryForReading("spectra", options["declevel"])
    if not check:
        # probably because this decimation level not calculated
        projectError(
            "Spectra file does not exist at level {}".format(options["declevel"]),
            quitRun=True,
        )
    return specReader


def calculateSpectra(projData: ProjectData, **kwargs) -> None:
    """Calculate spectra for the project time data

    The philosophy is that spectra are calculated out for all data and later limited using statistics and time constraints

    Parameters
    ----------
    projData : ProjectData
        A project data object
    sites : str, List[str], optional
        Either a single site or a list of sites
    sampleFreqs : int, float, List[float], optional
        The frequencies in Hz for which to calculate the spectra. Either a single frequency or a list of them.
    chans : List[str], optional
        The channels for which to calculate out the spectra
    polreverse :  Dict[str, bool]
        Keys are channels and values are boolean flags for reversing
    calibrate : bool, optional
        Flag whether to calibrate the data or not
    notch : List[float], optional
        List of frequencies to notch
    filter : Dict, optional
        Filter parameters
    specdir : str, optional
        The spectra directory to save the spectra data in
    """

    # default options
    options = {}
    options["sites"] = projData.getSites()
    options["sampleFreqs"]: List[float] = projData.getSampleFreqs()
    options["chans"]: List[str] = []
    options["polreverse"]: Union[bool, Dict[str, bool]] = False    
    options["calibrate"]: bool = True
    options["notch"]: List[float] = []
    options["filter"]: Dict = {}
    options["specdir"]: str = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    # prepare calibrator
    cal = getCalibrator(projData.calPath, projData.config)
    if options["calibrate"]:
        cal.printInfo()

    datetimeRef = projData.refTime
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        siteData.printInfo()

        # calculate spectra for each frequency
        for sampleFreq in options["sampleFreqs"]:
            measurements = siteData.getMeasurements(sampleFreq)
            projectText(
                "Site {} has {:d} measurement(s) at sampling frequency {:.2f}".format(
                    site, len(measurements), sampleFreq
                )
            )
            if len(measurements) == 0:
                continue  # no data files at this sample rate

            for meas in measurements:
                projectText(
                    "Calculating spectra for site {} and measurement {}".format(
                        site, meas
                    )
                )
                # get measurement start and end times - this is the time of the first and last sample
                reader = siteData.getMeasurement(meas)
                startTime = siteData.getMeasurementStart(meas)
                stopTime = siteData.getMeasurementEnd(meas)
                dataChans = (
                    options["chans"]
                    if len(options["chans"]) > 0
                    else reader.getChannels()
                )
                timeData = reader.getPhysicalData(startTime, stopTime, chans=dataChans)
                timeData.addComment(breakComment())
                timeData.addComment("Calculating project spectra")
                timeData.addComment(projData.config.getConfigComment())

                # apply various options
                applyPolarisationReversalOptions(options, timeData)
                applyCalibrationOptions(options, cal, timeData, reader)
                applyFilterOptions(options, timeData)
                applyNotchOptions(options, timeData)

                # define decimation and window parameters
                decParams = getDecimationParameters(sampleFreq, projData.config)
                decParams.printInfo()
                numLevels = decParams.numLevels
                winParams = getWindowParameters(decParams, projData.config)
                dec = Decimator(timeData, decParams)
                timeData.addComment(
                    "Decimating with {} levels and {} frequencies per level".format(
                        numLevels, decParams.freqPerLevel
                    )
                )

                # loop through decimation levels
                for iDec in range(0, numLevels):
                    # get the data for the current level
                    check = dec.incrementLevel()
                    if not check:
                        break  # not enough data
                    timeData = dec.timeData

                    # create the windower and give it window parameters for current level
                    sampleFreqDec = dec.sampleFreq
                    win = Windower(
                        datetimeRef,
                        timeData,
                        winParams.getWindowSize(iDec),
                        winParams.getOverlap(iDec),
                    )
                    if win.numWindows < 2:
                        break  # do no more decimation

                    # add some comments
                    timeData.addComment(
                        "Evaluation frequencies for this level {}".format(
                            listToString(decParams.getEvalFrequenciesForLevel(iDec))
                        )
                    )
                    timeData.addComment(
                        "Windowing with window size {} samples and overlap {} samples".format(
                            winParams.getWindowSize(iDec), winParams.getOverlap(iDec)
                        )
                    )

                    # create the spectrum calculator and statistics calculators
                    specCalc = SpectrumCalculator(
                        sampleFreqDec, winParams.getWindowSize(iDec)
                    )
                    # get ready a file to save the spectra
                    specPath = os.path.join(
                        siteData.getMeasurementSpecPath(meas), options["specdir"]
                    )
                    specWrite = SpectrumWriter(specPath, datetimeRef)
                    specWrite.openBinaryForWriting(
                        "spectra",
                        iDec,
                        sampleFreqDec,
                        winParams.getWindowSize(iDec),
                        winParams.getOverlap(iDec),
                        win.winOffset,
                        win.numWindows,
                        dataChans,
                    )

                    # loop though windows, calculate spectra and save
                    for iW in range(0, win.numWindows):
                        # get the window data
                        winData = win.getData(iW)
                        # calculate spectra
                        specData = specCalc.calcFourierCoeff(winData)
                        # write out spectra
                        specWrite.writeBinary(specData)

                    # close spectra file
                    specWrite.writeCommentsFile(timeData.getComments())
                    specWrite.closeFile()


def viewSpectra(
    projData: ProjectData, site: str, meas: str, **kwargs
) -> Union[plt.figure, None]:
    """View spectra for a measurement

    Parameters
    ----------
    projData : projecData
        The project data
    site : str
        The site to view
    meas: str
        The measurement of the site to view    
    chans : List[str], optional
        Channels to plot
    declevel : int, optional
        Decimation level to plot
    plotwindow : int, str, Dict, optional
        Windows to plot (local). If int, the window with local index plotwindow will be plotted. If string and "all", all the windows will be plotted if there are less than 20 windows, otherwise 20 windows throughout the whole spectra dataset will be plotted. If a dictionary, needs to have start and stop to define a range.
    specdir : str, optional
        String that specifies spectra directory for the measurement
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict, optional
        Dictionary of plot options
    
    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None and the figure is closed.
    """

    options = {}
    options["chans"]: List[str] = []
    options["declevel"]: int = 0
    options["plotwindow"]: Union[int, Dict, str] = [0]
    options["specdir"]: str = projData.config.configParams["Spectra"]["specdir"]
    options["show"]: bool = True
    options["save"]: bool = False
    options["plotoptions"]: Dict = plotOptionsSpec()
    options = parseKeywords(options, kwargs)

    projectText("Plotting spectra for measurement {} and site {}".format(meas, site))
    specReader = getSpecReader(projData, site, meas, **options)

    # channels
    dataChans = specReader.getChannels()
    if len(options["chans"]) > 0:
        dataChans = options["chans"]
    numChans = len(dataChans)

    # get windows
    numWindows = specReader.getNumWindows()
    sampleFreqDec = specReader.getSampleFreq()

    # get the window data
    windows = options["plotwindow"]
    if isinstance(windows, str) and windows == "all":
        if numWindows > 20:
            windows = list(
                np.linspace(0, numWindows, 20, endpoint=False, dtype=np.int32)
            )
        else:
            windows = list(np.arange(0, numWindows))
    elif isinstance(windows, int):
        windows = [windows]  # if an integer, make it into a list
    elif isinstance(windows, dict):
        windows = list(np.arange(windows["start"], windows["stop"] + 1))

    # create a figure
    plotfonts = options["plotoptions"]["plotfonts"]
    cmap = colorbarMultiline()
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])
    for iW in windows:
        if iW >= numWindows:
            break
        color = cmap(iW/numWindows)
        winData = specReader.readBinaryWindowLocal(iW)
        winData.view(
            fig=fig,
            chans=dataChans,
            label="{} to {}".format(
                winData.startTime.strftime("%m-%d %H:%M:%S"),
                winData.stopTime.strftime("%m-%d %H:%M:%S"),
            ),
            plotfonts=plotfonts,
            color=color,
        )

    st = fig.suptitle(
        "Spectra plot, site = {}, meas = {}, fs = {:.2f} [Hz], decimation level = {:2d}".format(
            site, meas, sampleFreqDec, options["declevel"]
        ),
        fontsize=plotfonts["suptitle"],
    )
    st.set_y(0.98)

    # put on axis labels etc
    for idx, chan in enumerate(dataChans):
        ax = plt.subplot(numChans, 1, idx + 1)
        plt.title("Amplitude {}".format(chan), fontsize=plotfonts["title"])
        if len(options["plotoptions"]["amplim"]) == 2:
            ax.set_ylim(options["plotoptions"]["amplim"])
        ax.set_xlim(0, specReader.getSampleFreq() / 2.0)
        plt.grid(True)

    # fig legend and formatting
    ax = plt.gca()
    h, l = ax.get_legend_handles_labels()
    fig.tight_layout(rect=[0.02, 0.02, 0.77, 0.92])
    # legend axis
    legax = plt.axes(position=[0.77, 0.02, 0.23, 0.88], in_layout=False)
    plt.tick_params(left=False, labelleft=False, bottom=False, labelbottom="False")
    plt.box(False)
    legax.legend(h, l, loc="upper left", fontsize=plotfonts["legend"])

    # plot show and save
    if options["save"]:
        impath = projData.imagePath
        filename = "spectraData_{}_{}_dec{}_{}".format(
            site, meas, options["declevel"], options["specdir"]
        )
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None
    return fig


def viewSpectraSection(
    projData: ProjectData, site: str, meas: str, **kwargs
) -> Union[plt.figure, None]:
    """View spectra section for a measurement

    Parameters
    ----------
    projData : projecData
        The project data
    site : str
        The site to view
    meas: str
        The measurement of the site to view    
    chans : List[str], optional
        Channels to plot
    declevel : int, optional
        Decimation level to plot
    specdir : str, optional
        String that specifies spectra directory for the measurement
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict, optional
        Dictionary of plot options
    
    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None and the figure is closed.
    """

    options = {}
    options["chans"] = []
    options["declevel"] = 0
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["show"] = True
    options["save"] = False
    options["plotoptions"] = plotOptionsSpec()
    options = parseKeywords(options, kwargs)

    projectText(
        "Plotting spectra section for measurement {} and site {}".format(meas, site)
    )
    specReader = getSpecReader(projData, site, meas, **options)

    # channels
    dataChans = specReader.getChannels()
    if len(options["chans"]) > 0:
        dataChans = options["chans"]

    # get windows
    numWindows = specReader.getNumWindows()
    sampleFreqDec = specReader.getSampleFreq()

    # freq array
    f = specReader.getFrequencyArray()

    # now if plotting a section, ignore plotwindow for now
    if numWindows > 250:
        windows = list(np.linspace(0, numWindows, 250, endpoint=False, dtype=np.int32))
    else:
        windows = np.arange(0, 250)

    # create figure
    plotfonts = options["plotoptions"]["plotfonts"]
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])
    st = fig.suptitle(
        "Spectra section, site = {}, meas = {}, fs = {:.2f} [Hz], decimation level = {:2d}, windows = {:d}, {} to {}".format(
            site,
            meas,
            sampleFreqDec,
            options["declevel"],
            len(windows),
            windows[0],
            windows[-1],
        ),
        fontsize=plotfonts["suptitle"],
    )
    st.set_y(0.98)

    # collect the data
    specData = np.empty(
        shape=(len(windows), len(dataChans), specReader.getDataSize()), dtype="complex"
    )
    dates = []
    for idx, iW in enumerate(windows):
        winData = specReader.readBinaryWindowLocal(iW)
        for cIdx, chan in enumerate(dataChans):
            specData[idx, cIdx, :] = winData.data[chan]
        dates.append(winData.startTime)

    ampLim = options["plotoptions"]["amplim"]
    for idx, chan in enumerate(dataChans):
        ax = plt.subplot(1, len(dataChans), idx + 1)
        plotData = np.transpose(np.absolute(np.squeeze(specData[:, idx, :])))
        if len(ampLim) == 2:
            plt.pcolor(
                dates,
                f,
                plotData,
                norm=LogNorm(vmin=ampLim[0], vmax=ampLim[1]),
                cmap=colorbar2dSpectra(),
            )
        else:
            plt.pcolor(
                dates,
                f,
                plotData,
                norm=LogNorm(vmin=plotData.min(), vmax=plotData.max()),
                cmap=colorbar2dSpectra(),
            )
        cb = plt.colorbar()
        cb.ax.tick_params(labelsize=plotfonts["axisTicks"])
        # set axis limits
        ax.set_ylim(0, specReader.getSampleFreq() / 2.0)
        ax.set_xlim([dates[0], dates[-1]])
        if isMagnetic(chan):
            plt.title("Amplitude {} [nT]".format(chan), fontsize=plotfonts["title"])
        else:
            plt.title("Amplitude {} [mV/km]".format(chan), fontsize=plotfonts["title"])
        ax.set_ylabel("Frequency [Hz]", fontsize=plotfonts["axisLabel"])
        ax.set_xlabel("Time", fontsize=plotfonts["axisLabel"])
        # set tick sizes
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(plotfonts["axisTicks"])
        plt.grid(True)

    # plot format
    fig.autofmt_xdate(rotation=90, ha="center")
    fig.tight_layout(rect=[0.02, 0.02, 0.96, 0.92])

    # plot show and save
    if options["save"]:
        impath = projData.imagePath
        filename = "spectraSection_{}_{}_dec{}_{}".format(
            site, meas, options["declevel"], options["specdir"]
        )
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None
    return fig


def viewSpectraStack(
    projData: ProjectData, site: str, meas: str, **kwargs
) -> Union[plt.figure, None]:
    """View spectra stacks for a measurement

    Parameters
    ----------
    projData : projecData
        The project data
    site : str
        The site to view
    meas: str
        The measurement of the site to view
    chans : List[str], optional
        Channels to plot
    declevel : int, optional
        Decimation level to plot
    numstacks : int, optional
        The number of windows to stack
    coherences : List[List[str]], optional
        A list of coherences to add, specified as [["Ex", "Hy"], ["Ey", "Hx"]] 
    specdir : str, optional
        String that specifies spectra directory for the measurement
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict, optional
        Dictionary of plot options
    
    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None and the figure is closed.
    """

    options = {}
    options["chans"] = []
    options["declevel"] = 0
    options["numstacks"] = 10
    options["coherences"] = []
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["show"] = True
    options["save"] = False
    options["plotoptions"] = plotOptionsSpec()
    options = parseKeywords(options, kwargs)

    projectText(
        "Plotting spectra stack for measurement {} and site {}".format(meas, site)
    )
    specReader = getSpecReader(projData, site, meas, **options)

    # channels
    dataChans = specReader.getChannels()
    if len(options["chans"]) > 0:
        dataChans = options["chans"]
    numChans = len(dataChans)

    # get windows
    numWindows = specReader.getNumWindows()
    sampleFreqDec = specReader.getSampleFreq()
    f = specReader.getFrequencyArray()

    # calculate num of windows to stack in each set
    stackSize = int(np.floor(1.0 * numWindows / options["numstacks"]))

    # calculate number of rows - in case interested in coherences too
    nrows = (
        2
        if len(options["coherences"]) == 0
        else 2 + np.ceil(1.0 * len(options["coherences"]) / numChans)
    )

    # setup the figure
    plotfonts = options["plotoptions"]["plotfonts"]
    cmap = colorbarMultiline()
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])
    st = fig.suptitle(
        "Spectra stack, fs = {:.6f} [Hz], decimation level = {:2d}, windows in each set = {:d}".format(
            sampleFreqDec, options["declevel"], stackSize
        ),
        fontsize=plotfonts["suptitle"],
    )
    st.set_y(0.98)

    # do the stacking
    for iP in range(0, options["numstacks"]):
        stackStart = iP * stackSize
        stackStop = min(stackStart + stackSize, numWindows)
        color = cmap(iP/options["numstacks"])
        # dictionaries to hold data for this section
        stackedData = {}
        ampData = {}
        phaseData = {}
        powerData = {}

        # assign initial zeros
        for c in dataChans:
            stackedData[c] = np.zeros(shape=(specReader.getDataSize()), dtype="complex")
            ampData[c] = np.zeros(shape=(specReader.getDataSize()), dtype="complex")
            phaseData[c] = np.zeros(shape=(specReader.getDataSize()), dtype="complex")
            for c2 in dataChans:
                powerData[c + c2] = np.zeros(
                    shape=(specReader.getDataSize()), dtype="complex"
                )

        # now stack the data and create nice plots
        for iW in range(stackStart, stackStop):
            winData = specReader.readBinaryWindowLocal(iW)
            for c in dataChans:
                stackedData[c] += winData.data[c]
                ampData[c] += np.absolute(winData.data[c])
                phaseData[c] += np.angle(winData.data[c]) * (180.0 / np.pi)
                # get coherency data
                for c2 in dataChans:
                    powerData[c + c2] += winData.data[c] * np.conjugate(
                        winData.data[c2]
                    )
            if iW == stackStart:
                startTime = winData.startTime
            if iW == stackStop - 1:
                stopTime = winData.stopTime

        # scale powers and stacks
        ampLim = options["plotoptions"]["amplim"]
        for idx, c in enumerate(dataChans):
            stackedData[c] = stackedData[c] / (stackStop - stackStart)
            ampData[c] = ampData[c] / (stackStop - stackStart)
            phaseData[c] = phaseData[c] / (stackStop - stackStart)
            for c2 in dataChans:
                # normalisation
                powerData[c + c2] = 2 * powerData[c + c2] / (stackStop - stackStart)
                # normalisation
                powerData[c + c2][[0, -1]] = powerData[c + c2][[0, -1]] / 2

            # plot
            ax1 = plt.subplot(nrows, numChans, idx + 1)
            plt.title("Amplitude {}".format(c), fontsize=plotfonts["title"])
            h = ax1.semilogy(
                f,
                ampData[c],
                color=color,
                label="{} to {}".format(
                    startTime.strftime("%m-%d %H:%M:%S"),
                    stopTime.strftime("%m-%d %H:%M:%S"),
                ),
            )
            if len(ampLim) > 2:
                ax1.set_ylim(ampLim)
            else:
                ax1.set_ylim(0.01, 1000)
            ax1.set_xlim(0, sampleFreqDec / 2.0)
            if isMagnetic(c):
                ax1.set_ylabel("Amplitude [nT]", fontsize=plotfonts["axisLabel"])
            else:
                ax1.set_ylabel("Amplitude [mV/km]", fontsize=plotfonts["axisLabel"])
            ax1.set_xlabel("Frequency [Hz]", fontsize=plotfonts["axisLabel"])
            plt.grid(True)

            # set tick sizes
            for label in ax1.get_xticklabels() + ax1.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])
            # plot phase
            ax2 = plt.subplot(nrows, numChans, numChans + idx + 1)
            plt.title("Phase {}".format(c), fontsize=plotfonts["title"])
            ax2.plot(
                f,
                phaseData[c],
                color=color,                
                label="{} to {}".format(
                    startTime.strftime("%m-%d %H:%M:%S"),
                    stopTime.strftime("%m-%d %H:%M:%S"),
                ),
            )
            ax2.set_ylim(-180, 180)
            ax2.set_xlim(0, sampleFreqDec / 2.0)
            ax2.set_ylabel("Phase [degrees]", fontsize=plotfonts["axisLabel"])
            ax2.set_xlabel("Frequency [Hz]", fontsize=plotfonts["axisLabel"])
            plt.grid(True)
            # set tick sizes
            for label in ax2.get_xticklabels() + ax2.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])

        # plot coherences
        for idx, coh in enumerate(options["coherences"]):
            c = coh[0]
            c2 = coh[1]
            cohNom = np.power(np.absolute(powerData[c + c2]), 2)
            cohDenom = powerData[c + c] * powerData[c2 + c2]
            coherence = cohNom / cohDenom
            ax = plt.subplot(nrows, numChans, 2 * numChans + idx + 1)
            plt.title("Coherence {} - {}".format(c, c2), fontsize=plotfonts["title"])
            ax.plot(
                f,
                coherence,
                color=color,
                label="{} to {}".format(
                    startTime.strftime("%m-%d %H:%M:%S"),
                    stopTime.strftime("%m-%d %H:%M:%S"),
                ),
            )
            ax.set_ylim(0, 1.1)
            ax.set_xlim(0, sampleFreqDec / 2)
            ax.set_ylabel("Coherence", fontsize=plotfonts["axisLabel"])
            ax.set_xlabel("Frequency [Hz]", fontsize=plotfonts["axisLabel"])
            plt.grid(True)
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])

    # fig legend and layout
    ax = plt.gca()
    h, l = ax.get_legend_handles_labels()
    fig.tight_layout(rect=[0.01, 0.01, 0.98, 0.81])
    # legend
    legax = plt.axes(position=[0.01, 0.82, 0.98, 0.12], in_layout=False)
    plt.tick_params(left=False, labelleft=False, bottom=False, labelbottom="False")
    plt.box(False)
    legax.legend(h, l, ncol=4, loc="upper center", fontsize=plotfonts["legend"])

    # plot show and save
    if options["save"]:
        impath = projData.imagePath
        filename = "spectraStack_{}_{}_dec{}_{}".format(
            site, meas, options["declevel"], options["specdir"]
        )
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None
    return fig



import sys
import os
import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
from typing import List, Dict, Union

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsPrint import arrayToString
from resistics.utilities.utilsIO import getDataDirsInDirectory
from resistics.utilities.utilsReader import getDataReader
from resistics.utilities.utilsPlotter import getViewFonts


class SiteData(DataObject):
    """Class for holding site information

    Attributes
    ----------
    timePath : int
        Path to site time data
    specPath : float
        Path to site spectra data
    statPath : datetime.datetime
        Path to site stat data
    transFuncPath : datetime.datetime
        Path to site transfer function data 
    measurements : List[str]
        A list of the measurements in the site
    spectra : List[str]
        A list of spectra data for the site
    statistics : List[str]
        A list of statistics data for the site
    transferFunctions : List[str]
        A list of transfer function data for the site
    siteStart : datetime.datetime
        The time of the first sample in the site
    siteEnd : datetime.datetime
        The time of the last sample in the site
    readers : Dict
        Data readers for each measurement
    starts : Dict

    Methods
    -------
    __init__(siteName, timePath, specPath, statPath, transFuncPath)
        Initialise the site data
    refresh()
        Refresh site data
    getSampleFreqs()
        Get list of unique sampling frequencies in site
    getMeasurements(fs=None) 
        Get list of measurements in site, optionally filtered by sampling frequency
    getMeasurement(meas)
        Get the dataReader for the measurement
    getMeasurementSampleFreq(meas)
        Get the sampling frequency of a measurement in Hz        
    getMeasurementTimePath(meas)
        Get the path for a measurement
    getMeasurementSpecPath(meas)
        Get the path for the spectra of a measurement
    getMeasurementStatPath(meas)
        Get the path for the statistics of a measurement
    getMeasurementTransFuncPath(meas)
        Get the path for the transfer function data of a measurement
    getMeasurementStart(meas)
        Get the time of the first sample of a measurement
    getMeasurementEnd(meas)
        Get the time of the last sample of a measurement
    view(kwargs)
        View the site data measurement information 
    printList()
        Class status returned as list of strings          
    """

    def __init__(
        self,
        siteName: str,
        timePath: str,
        specPath: str,
        statPath: str,
        maskPath: str,
        transFuncPath: str,
    ) -> None:
        self.siteName: str = siteName
        self.timePath: str = os.path.join(timePath, siteName)
        self.specPath: str = os.path.join(specPath, siteName)
        self.statPath: str = os.path.join(statPath, siteName)
        self.maskPath: str = os.path.join(maskPath, siteName)
        self.transFuncPath: str = os.path.join(transFuncPath, siteName)
        # measurement information
        self.measurements: List[str] = []
        self.spectra: List[str] = []
        self.statistics: List[str] = []
        self.transferFunctions: List[str] = []
        self.siteStart = datetime.now()
        self.siteEnd = datetime.now()
        self.readers: Dict = {}
        self.starts: Dict[str, datetime] = {}
        self.ends: Dict[str, datetime] = {}
        self.fs: Dict[str, float] = {}
        # refresh site data
        self.refresh()

    def refresh(self):
        """ Refresh site information"""

        self.measurements = getDataDirsInDirectory(self.timePath)
        for meas in self.measurements:
            # initialise datapath and the various readers
            datapath = os.path.join(self.timePath, meas)
            reader = getDataReader(datapath)
            if not reader:  # then no data found in this directory
                self.printWarning(
                    "No data found in measurement directory {} for site {}. Please remove this directory".format(
                        meas, self.siteName
                    )
                )
                continue
            # otherwise get the reader and information
            self.readers[meas] = reader
            self.starts[meas] = reader.getStartDatetime()
            self.ends[meas] = reader.getStopDatetime()
            self.fs[meas] = reader.getSampleFreq()

        # start and end time of project
        if len(self.measurements) != 0:
            self.siteStart = min(self.starts.values())
            self.siteEnd = max(self.ends.values())

        # spectra, statistic and transfer function data
        self.spectra = getDataDirsInDirectory(self.specPath)
        self.statistics = getDataDirsInDirectory(self.statPath)
        self.transferFunctions = getDataDirsInDirectory(self.transFuncPath)

    def getSampleFreqs(self) -> List[float]:
        """Get a list of all the distinct sampling frequencies in the site

        Returns
        -------
        List[float]
            List of the unique sampling frequencies in a site
        """

        # convert to set to get only unique values
        sampleFreqs = set(self.fs.values())
        return sorted(list(sampleFreqs))

    def getMeasurements(self, sampleFreq: Union[None, float, int] = None) -> List[str]:
        """Get a list of measurements for a site

        Parameters
        ----------
        sampleFreq : float, int, optional
            Sampling frequency

        Returns
        -------
        List[str]
            List of site time files. If fs is supplied, then only time files with sampling frequency fs.
        """

        if sampleFreq is None:
            return self.measurements
        # fs is supplied
        filteredMeasurements = []
        for meas in self.measurements:
            if self.fs[meas] == sampleFreq:
                filteredMeasurements.append(meas)
        return filteredMeasurements

    def getMeasurement(self, meas: str):
        """Get the data reader of a measurement at a site

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        DataReader, bool
            Returns the data reader for the measurement or if the measurement is not found, False
        """

        if self.checkMeasurement(meas):
            return self.readers[meas]
        self.printWarning(
            "Meausrement directory {} for site {} not found".format(self.siteName, meas)
        )
        return False

    def getMeasurementSampleFreq(self, meas: str) -> float:
        """Get the sample frequency in Hz of a particular measurement at a site

        Parameters
        ----------
        site : str
            The name of the site
        meas : str
            The name of the measurement

        Returns
        -------
        float
            Sampling frequency of site
        """

        return self.fs[meas]

    def getMeasurementTimePath(self, meas: str) -> str:
        """Get the time data path for a measurement at a site

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        str
            Path to the time data for a measurement at the site
        """

        return os.path.join(self.timePath, meas)

    def getMeasurementSpecPath(self, meas: str) -> str:
        """Get the spectra data path for a measurement at a site

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        str
            Path to the spectra data for a measurement at the site
        """

        return os.path.join(self.specPath, meas)

    def getMeasurementStatPath(self, meas: str) -> str:
        """Get the statistic data path for a measurement at a site

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        str
            Path to the statistic data for a measurement at the site
        """

        return os.path.join(self.statPath, meas)

    def getSpecdirMaskPath(self, specdir: str) -> str:
        """Get the mask path for a particular set of spectra calculations

        Masks are calculated with relation to sampling frequencies over a whole set of spectra directories.

        Parameters
        ----------
        specdir : str
            The spectra set

        Returns
        -------
        str
            Path to mask data for the spectra set
        """

        return os.path.join(self.maskPath, specdir)

    def getMeasurementTransFuncPath(self, meas: str) -> str:
        """Get the transfer function data path for a measurement at a site

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        str
            Path to the transfer function data for a measurement at the site
        """

        return os.path.join(self.transFuncPath, meas)

    def getMeasurementStart(self, meas: str) -> datetime:
        """Get the start time of a particular measurement at a site

        This returns the time of the first sample for the measurement

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        datetime.datetime
            Start time of a measurement at a site
        """

        return self.starts[meas]

    def getMeasurementEnd(self, meas: str) -> datetime:
        """Get the end time of a particular measurement at a site

        This returns the time of the last sample for the measurement

        Parameters
        ----------
        meas : str
            The name of the measurement

        Returns
        -------
        datetime.datetime
            End time of a measurement at a site
        """

        return self.ends[meas]

    def checkMeasurement(self, meas: str) -> bool:
        """Check if site and measurement are part of the project
        
        Parameters
        ----------
        meas : str
            Name of measurement

        Returns
        -------
        bool 
            True if measurement exists
        """

        if meas in self.measurements:
            return True
        else:
            return False

    def view(self, **kwargs) -> plt.figure:
        """Plot a timeline of the measurements in the site
        
        Parameters
        ----------
        figsize : Tuple (width, height), optional
            The figure size
        plotfonts : Dict, optional
            Fonts to use for plotting
        show : bool, optional
            Boolean flag for showing
        """

        import matplotlib.dates as mdates
        from matplotlib.patches import Rectangle
        import numpy as np
        from resistics.utilities.utilsPlotter import getPlotFonts

        figsize = kwargs["figsize"] if "figsize" in kwargs else (15, 8)
        plotFonts = kwargs["plotFonts"] if "plotFonts" in kwargs else getPlotFonts()
        show = kwargs["show"] if "show" in kwargs else True

        fig = plt.figure(figsize=figsize)
        ax = plt.subplot(1, 1, 1)
        # get xlimits
        xStart = self.siteStart - timedelta(days=2)
        xEnd = self.siteEnd + timedelta(days=2)

        numMeas = len(self.measurements)
        height = 0.4
        for idx, meas in enumerate(self.measurements):
            idx = numMeas - idx
            measStart = mdates.date2num(self.starts[meas])
            measEnd = mdates.date2num(self.ends[meas])
            measWidth = measEnd - measStart
            yOffset = 0.5 + height
            rect = Rectangle((measStart, idx - yOffset), measWidth, 2 * height)
            ax.add_patch(rect)
        # x axis formatting
        days = mdates.DayLocator()
        daysFmt = mdates.DateFormatter("%Y-%m-%d")
        ax.xaxis.set_major_locator(days)
        ax.xaxis.set_major_formatter(daysFmt)
        plt.xlim([xStart, xEnd])
        ax.xaxis.grid(True, which="major", ls="--", color="gainsboro")
        plt.xlabel("Date", fontsize=plotFonts["axisLabel"])
        fig.autofmt_xdate(rotation=30)
        for label in ax.get_xticklabels():
            label.set_fontsize(plotFonts["axisTicks"])           
        # yaxis formatting
        plt.ylim([0, numMeas])
        yticks = np.arange(1, numMeas + 1)
        ax.set_yticks(yticks)
        ax.set_yticklabels([])
        ax.set_yticks(yticks - 0.5, minor=True)
        ytickLabels = []
        for meas in self.measurements:
            ytickLabels.append("{}\n{:.1f} Hz".format(meas, self.fs[meas]))
        ax.set_yticklabels(
            ytickLabels[::-1], minor=True, fontsize=plotFonts["axisTicks"]
        )
        ax.yaxis.grid(True, which="major", ls="--", color="gainsboro")
        plt.ylabel("Measurements", fontsize=plotFonts["axisLabel"], labelpad=20)
        # title
        plt.title("Site timeline", fontsize=plotFonts["title"])

        fig.tight_layout()
        if show:
            plt.show()
        return fig

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Site = {}".format(self.siteName))
        textLst.append("Time data path = {}".format(self.timePath))
        textLst.append("Spectra data path = {}".format(self.specPath))
        textLst.append("Statistics data path = {}".format(self.statPath))
        textLst.append("TransFunc data path = {}".format(self.transFuncPath))
        textLst.append("Site start time = {}".format(self.siteStart))
        textLst.append("Site stop time = {}".format(self.siteEnd))
        sampleFreqText = (
            "No measurement files"
            if len(self.measurements) == 0
            else arrayToString(self.getSampleFreqs())
        )
        textLst.append("Sampling frequencies recorded = {}".format(sampleFreqText))

        # print any measurement directors
        textLst.append(
            "Number of measurement files = {}".format(len(self.measurements))
        )
        if len(self.measurements) > 0:
            textLst.append(
                "Measurement\t\tSample Frequency (Hz)\t\tStart Time\t\tEnd Time"
            )
            for meas in self.measurements:
                textLst.append(
                    "{}\t\t{}\t\t{}\t\t{}".format(
                        meas, self.fs[meas], self.starts[meas], self.ends[meas]
                    )
                )
        else:
            textLst.append("No time data files (measurement files) found")
        return textLst



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from datetime import datetime, timedelta
from typing import List, Dict, Any
from configobj import ConfigObj

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsPrint import listToString


class ConfigData(DataObject):
    """Class for holding cofiguration options

    Attributes
    ----------
    configFile : str
        Path to the configuration file
    configParams : ConfigObj
        ConfigObj object to parse configuration information

    Methods
    -------
    __init__(configFile)
        Initialise configuration data
    configure()
        Configure all the options
    printList()
        Class information as a list of strings
    """

    def __init__(self, configFile: str = "") -> None:
        """Initialise with default configuration options
        
        Parameters
        ----------
        configFile : str
            Path to the configuration file
        """

        self.configFile: str = configFile
        self.configParams = loadConfig(self.configFile)
        # options dictionary
        self.flags: Dict[str, bool] = {}
        self.configure()

    def configure(self) -> None:
        """Set the configuration parameters
        
        This sets a couple of flags in the configuration to regarding whether a custom set of evaluation frequencies are being used or a custom window size definition
        """

        # frequencies
        self.flags["customfrequencies"] = False
        if len(self.configParams["Frequencies"]["frequencies"]) > 0:
            # custom frequencies have been provided
            self.flags["customfrequencies"] = True
            # check to make sure appropriate number of frequencies provided
            numUserFreqs = len(self.configParams["Frequencies"]["frequencies"])
            numExpectedFreqs = (
                self.configParams["Decimation"]["numlevels"]
                * self.configParams["Frequencies"]["perlevel"]
            )
            if numUserFreqs != numExpectedFreqs:
                self.printError(
                    "Custom frequencies provided. Please ensure that the number of frequencies is equal to numlevels * perlevel",
                    quitRun=True,
                )

        # windows
        self.flags["customwindows"] = False
        numUserWindows = len(self.configParams["Window"]["windowsizes"])
        numUserOverlaps = len(self.configParams["Window"]["overlapsizes"])
        if numUserWindows > 0 and numUserOverlaps > 0:
            self.flags["customwindows"] = True
            numlevels = self.configParams["Decimation"]["numlevels"]
            if numUserWindows != numlevels or numUserOverlaps != numlevels:
                self.printError(
                    "When supplying custom window sizes, the number of window sizes provided should be equal to the number of decimation levels (numlevels)",
                    quitRun=True,
                )

    def getConfigComment(self) -> str:
        """Returns a string to add as a comment to data

        Returns
        -------
        str 
            Comment string to add to data
        """

        if self.configFile == "":
            return "Using default configuration"
        else:
            return "Using configuration with name {} in configuration file {}".format(self.configParams["name"], self.configFile)

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        if self.configFile == "":
            textLst.append("Configuration file = Default configuration")
        else:
            textLst.append("Configuration file = {:s}".format(self.configFile))
        
        textLst.append("Configuration name = {:s}".format(self.configParams["name"]))

        textLst.append("Flags:")
        for key, value in self.flags.items():
            textLst.append("{:s} = {}".format(key, value))

        textLst.append("Configuration Parameters:")
        textLst.append("Name = {:s}".format(self.configParams["name"]))
        textLst = textLst + self.printListSection("Calibration")
        textLst = textLst + self.printListSection("Decimation")
        textLst = textLst + self.printListSection("Frequencies")
        textLst = textLst + self.printListSection("Window")
        textLst = textLst + self.printListSection("Spectra")
        textLst = textLst + self.printListSection("Statistics")
        textLst = textLst + self.printListSection("Solver")
        
        return textLst

    def printListSection(self, section: str) -> List[str]:
        """Configuration section information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("{:s}:".format(section))
        for key, value in self.configParams[section].items():
            textLst.append("\t{:s} = {}".format(key, value))
        defaultOptions = "No defaults used"
        if len(self.configParams[section].defaults) > 0:
            defaultOptions = listToString(self.configParams[section].defaults) 
        textLst.append("\tDefaulted options = {:s}".format(defaultOptions))
        
        return textLst